import sys

import click

from gen3_tracker.common import CLIOutput
from gen3_tracker.config import Config, ensure_auth
from gen3_tracker.gen3.buckets import get_buckets
from gen3_tracker.projects.lister import ls
from gen3_tracker.projects.remover import empty
from gen3_tracker import NaturalOrderGroup, ENV_VARIABLE_PREFIX


@click.group(name='projects', cls=NaturalOrderGroup)
@click.pass_obj
def project_group(config: Config):
    """Manage Gen3 projects."""
    pass


@project_group.command(name="ls")
@click.option('--verbose', default=False, show_default=True, is_flag=True,
              help="List all project details")
@click.pass_obj
def project_ls(config: Config, verbose: bool):
    """List all projects user has access to."""
    with CLIOutput(config=config) as output:
        try:
            auth = ensure_auth(config=config)
            output.update(ls(config, auth=auth, full=verbose))
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e


@project_group.command(name="empty")
@click.option('--project_id', default=None, show_default=True,
              help="Gen3 program-project", envvar=f"{ENV_VARIABLE_PREFIX}PROJECT_ID")
@click.option('--confirm', default=None, show_default=True,
              help="Enter 'empty' to confirm", envvar=f"{ENV_VARIABLE_PREFIX}PROJECT_ID")
@click.pass_obj
def project_empty(config: Config, project_id: str, confirm: str):
    """Empty all metadata (graph, flat) for a project."""
    with CLIOutput(config=config) as output:
        try:
            assert confirm == 'empty', "Please confirm by entering --confirm empty"
            if not project_id:
                project_id = config.gen3.project_id
                click.secho(f"No project_id provided, using current project {project_id}", fg='yellow',
                            file=sys.stderr)
            _ = empty(config, project_id)
            _['msg'] = f"Emptied {project_id}"
            output.update(_)
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e


@project_group.command(name="bucket")
@click.option('--project_id', default=None, show_default=True,
              help="Gen3 program-project", envvar=f"{ENV_VARIABLE_PREFIX}PROJECT_ID")
@click.pass_obj
def project_bucket(config: Config, project_id: str):
    """Show project bucket."""
    if not project_id:
        project_id = config.gen3.project_id
        click.secho(f"No project_id provided, using current project {project_id}", fg='yellow', file=sys.stderr)
    with CLIOutput(config=config) as output:
        try:
            program, project = project_id.split('-')
            buckets = get_buckets(config=config)
            for k, v in buckets['S3_BUCKETS'].items():
                assert 'programs' in v, f"no configured programs in fence buckets {v} {buckets}"
                if program in v['programs']:
                    output.update({k: v})
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e



import logging
import socket
from urllib.parse import urlparse

import requests
from gen3.auth import Gen3Auth
from gen3.index import Gen3Index

from gen3_tracker.git import DVC, DVCMeta


def write_indexd(auth: Gen3Auth,
                 project_id: str,
                 dvc: DVC,
                 bucket_name: str,
                 overwrite: bool,
                 restricted_project_id: str,
                 existing_records: list[str] = [],
                 legacy_bucket: bool = False,
                 message: str = None
                 ) -> bool:
    """Write manifest entry to indexd."""
    assert auth, "Expected auth"
    assert project_id, "Expected project_id"
    index_client = Gen3Index(auth)
    program, project = project_id.split('-')
    logger = logging.getLogger(__name__)
    dvc.project_id = project_id

    # SYNC
    existing_record = None
    hashes, metadata = create_hashes_metadata(dvc, program, project)

    if message:
        metadata['message'] = message

    if overwrite:
        existing_record = dvc.object_id in existing_records
        if existing_record:
            # SYNC
            # print(f"Deleting existing record {dvc.object_id}")
            index_client.delete_record(guid=dvc.object_id)
            existing_record = False

    authz = [f'/programs/{program}/projects/{project}']
    if restricted_project_id:
        _ = restricted_project_id.split('-')
        authz.append(f'/programs/{_[0]}/projects/{_[1]}')

    # strip any file:/// prefix
    dvc.out.path = urlparse(dvc.out.path).path

    # We need this for symlinked files
    if dvc.out.realpath:
        metadata['realpath'] = urlparse(dvc.out.realpath).path

    if not existing_record:
        try:
            file_name = dvc.out.path
            urls = [f"s3://{bucket_name}/{dvc.object_id}/{file_name}"]
            if legacy_bucket:
                urls = [f"s3://{bucket_name}/{file_name}"]
            if dvc.meta.no_bucket:
                hostname = socket.gethostname()
                _ = f"{hostname}/{metadata['realpath']}".replace('//', '/')
                urls = [f"scp://{_}"]
            if dvc.out.source_url:
                urls = [dvc.out.source_url]

            # print(f"Writing indexd record for {dvc.object_id} {urls}")
            response = index_client.create_record(
                did=dvc.object_id,
                hashes=hashes,
                size=dvc.out.size,
                authz=authz,
                file_name=file_name,
                metadata=metadata,
                urls=urls
            )
            assert response, "Expected response from indexd create_record"

        except (requests.exceptions.HTTPError, AssertionError) as e:
            if 'already exists' in str(e):
                logger.error(
                    f"indexd record already exists, consider using --overwrite. {dvc.object_id} {str(e)}")
            raise e
    return True


def create_hashes_metadata(dvc: DVC, program, project):
    meta: DVCMeta = dvc.meta

    if not meta:
        meta = DVCMeta()

    hashes = {dvc.out.hash: getattr(dvc.out, dvc.out.hash)}
    metadata = {
        **{
            'document_reference_id': dvc.object_id,
            'specimen_identifier': meta.specimen,
            'patient_identifier': meta.patient,
            'task_identifier': meta.task,
            'observation_identifier': meta.observation,
            'project_id': f'{program}-{project}',
            'no_bucket': meta.no_bucket,
        },
        **hashes}
    return hashes, metadata



from typing import List, Any, Union
from collections import defaultdict

from gen3.auth import Gen3Auth
from pydantic import BaseModel

from gen3_tracker.config import ensure_auth, Config


class ProjectSummary(BaseModel):
    """Summary of a project."""
    user_perms: bool = False
    """Project exists in sheepdog flag"""
    permissions: list[dict[str, Any]] = []


class ProjectSummaries(BaseModel):
    """Summary of projects, including messages."""
    endpoint: str
    """The commons url"""
    incomplete: Union[dict[str, ProjectSummary], list[str]] = {}
    """List of projects that require creation in sheepdog."""
    complete: Union[dict[str, ProjectSummary], list[str]] = {}
    """List of projects that exist in sheepdog."""
    messages: List[str] = []
    """List of messages"""


def get_user(config: Config = None, auth: Gen3Auth = None) -> dict:
    """Fetch information about the user."""
    if not auth:
        assert config
        auth = ensure_auth(config=config)
    return auth.curl('/user/user').json()


def recursive_defaultdict():
    return defaultdict(recursive_defaultdict)


def get_projects(auth) -> dict:
    """Return a dict of programs, projects and their existence flag."""

    # get the list of programs and projects from arborist
    user = get_user(auth=auth)
    arborist_projects = recursive_defaultdict()
    for _ in user['authz'].keys():
        if not all([_.startswith('/programs'), 'projects/' in _]):
            continue

        permissions = user['authz'][_]

        _ = _.replace('/programs/', '')
        _ = _.split('/')
        _program = _[0]
        _project = _[-1]
        arborist_projects[_program][_project]['permissions'] = permissions

        # Checking for all policies granted in reader and writer role in user.yaml
        has_read = any(_["method"] == "read" and _["service"] == "*" for _ in permissions)
        has_read_storage = any(_["method"] == "read-storage" and _["service"] == "*" for _ in permissions)
        has_create = any(_["method"] == "create" and _["service"] == "*" for _ in permissions)
        has_file_upload = any(_["method"] == "file_upload" and _["service"] == "fence" for _ in permissions)
        has_write_storage = any(_["method"] == "write-storage" and _["service"] == "*" for _ in permissions)
        has_update = any(_["method"] == "update" and _["service"] == "*" for _ in permissions)

        arborist_projects[_program][_project]['exists'] = False
        if all([has_read, has_read_storage, has_create, has_file_upload, has_write_storage, has_update]):
            arborist_projects[_program][_project]['exists'] = True

    return arborist_projects



import logging

from gen3.auth import Gen3Auth

from gen3_tracker.config import ensure_auth, Config


def get_buckets(config: Config = None, auth: Gen3Auth = None) -> dict:
    """Fetch information about the buckets."""
    if not auth:
        assert config
        auth = ensure_auth(config=config)

    response = auth.curl('/user/data/buckets')

    # TODO - remove when no longer needed
    if response.status_code == 405:
        logging.getLogger(__name__).warning(
            "TODO /data/buckets response returned 405, "
            "see https://cdis.slack.com/archives/CDDPLU1NU/p1683566639636949 "
            "see quay.io/cdis/fence:feature_bucket_info_endpoint "
        )

    assert response.status_code == 200, (response.status_code, response.content)
    return response.json()


def get_program_bucket(config: Config, program: str = None, auth: Gen3Auth = None) -> str:
    """Get the bucket for a program."""
    buckets = get_buckets(config=config, auth=auth)
    bucket_name = None
    if program is None:
        program = config.gen3.program

    for k, v in buckets['S3_BUCKETS'].items():
        assert 'programs' in v, f"no configured programs in fence buckets {v} {buckets}"
        if program in v['programs']:
            bucket_name = k
            break
    # assert bucket_name, f"could not find bucket for {program}"
    return bucket_name



import csv
import json
import logging
import os
import decimal
import pathlib
import re
import shutil
import sys
import uuid
import zipfile
from datetime import datetime
from hashlib import md5
from typing import Mapping, Iterator, Dict, TextIO, Generator
from urllib.parse import urlparse

import click
from dateutil import parser as dateutil_parser

import orjson
import yaml
from dateutil.tz import tzutc
from fhir.resources.documentreference import DocumentReference
from fhir.resources.identifier import Identifier
from pydantic import BaseModel
from pydantic.json import pydantic_encoder

import io
import gzip

from gen3_tracker import Config

# see https://github.com/uc-cdis/indexd/blob/1ae3fea6569e8765d5b7247123b66bde9ce4c39c/indexd/index/blueprint.py#L35-L42
ACCEPTABLE_HASHES = {
    "md5": re.compile(r"^[0-9a-f]{32}$").match,
    "sha1": re.compile(r"^[0-9a-f]{40}$").match,
    "sha256": re.compile(r"^[0-9a-f]{64}$").match,
    "sha512": re.compile(r"^[0-9a-f]{128}$").match,
    "crc": re.compile(r"^[0-9a-f]{8}$").match,
    "etag": re.compile(r"^[0-9a-f]{32}(-\d+)?$").match,
}

# Define color constants
INFO_COLOR = 'yellow'
ERROR_COLOR = 'red'
SUCCESS_COLOR = 'green'

PROJECT_DIR = '.g3t'
PROJECT_DIRECTORIES = [PROJECT_DIR, 'META/', 'MANIFEST/', 'logs/']
PROJECT_README = """
# Data Directory

Welcome to the data directory! This repository contains important data files for our project. Before you proceed, please take note of the following guidelines to ensure the security and integrity of our data.

## Important Note: Do Not Check in Protected Files

Some files in this directory are considered protected and contain sensitive information.

**DO NOT** check in or commit these protected files to the version control system (e.g., Git).

This is crucial to prevent unauthorized access and to comply with security and privacy policies.


## Usage Guidelines:

1. **Read-Only Access:** Unless you have explicit permission to modify or update the data, treat this directory as read-only.

2. **Data Integrity:** Ensure the integrity of the data by following proper procedures for reading, updating, and managing files.

3. **Security Awareness:** Be aware of the sensitivity of the data stored here and take necessary precautions to protect it from unauthorized access.


Thank you for your cooperation in maintaining the security and confidentiality of our data.
"""


def _default_json_serializer(obj):
    """JSON Serializer, render decimal and bytes types."""
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if isinstance(obj, bytes):
        return obj.decode()
    raise TypeError


def print_formatted(config, output: Mapping) -> None:
    """Print the output, using configured output format"""
    from gen3_tracker import Config
    config: Config = config
    if config.output.format == "yaml":
        print(yaml.dump(output, sort_keys=False))
    elif config.output.format == "json":
        print(
            orjson.dumps(
                output, default=pydantic_encoder, option=orjson.OPT_INDENT_2
            ).decode()
        )
    else:
        print(output)


def read_ndjson_file(path: str) -> Iterator[dict]:
    """Read ndjson file, load json line by line."""
    with _file_opener(path) as jsonfile:
        for l_ in jsonfile.readlines():
            yield orjson.loads(l_)


def read_json_file(path: str) -> Iterator[dict]:
    """Read ndjson file, load json line by line."""
    with _file_opener(path) as jsonfile:
        try:
            yield orjson.loads(jsonfile.read())
        except orjson.JSONDecodeError as e:
            logging.error(f"Error reading {path}: {e}")
            raise


def read_json(path: str) -> Iterator[dict]:
    """Read json or ndjson from file or zip."""
    if is_ndjson(path):
        _reader = read_ndjson_file
    else:
        _reader = read_json_file
    for _ in _reader(path):
        yield _


# TODO - unused, deprecate?
def read_tsv(path: str, delimiter="\t") -> Iterator[Dict]:
    """Read tsv file line by line."""
    with open(path) as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter=delimiter)
        for row in reader:
            yield row


def read_yaml(path: str) -> Dict:
    """Read a yaml file."""
    with open(path, "r") as fp:
        return yaml.safe_load(fp.read())


def is_url(to_) -> bool:
    """Does the destination parameter describe an upload? ie have an url.scheme"""
    return len(urlparse(to_).scheme) > 0


def is_json_extension(name: str) -> bool:
    """Files we are interested in"""
    if name.endswith('json.gz'):
        return True
    if name.endswith('json'):
        return True
    return False


def is_ndjson(file_path: pathlib.Path) -> bool:
    """Open file, check if ndjson."""
    fp = _file_opener(file_path)
    try:
        with fp:
            for line in fp.readlines():
                orjson.loads(line)
                break
        return True
    except Exception as e:  # noqa
        return False


def _file_opener(file_path):
    """Open file appropriately."""
    if isinstance(file_path, str):
        file_path = pathlib.Path(file_path)
    if file_path.name.endswith('gz'):
        fp = io.TextIOWrapper(io.BufferedReader(gzip.GzipFile(file_path)))  # noqa
    else:
        fp = open(file_path, "rb")
    return fp


class EmitterContextManager:
    """Maintain file pointers to output directory."""

    def __init__(self, output_path: str, verbose=False, file_mode="w",
                 logger=logging.getLogger("EmitterContextManager")):
        """Ensure output_path exists, init emitter dict."""
        output_path = pathlib.Path(output_path)
        if not output_path.exists():
            output_path.mkdir(parents=True)
        assert output_path.is_dir(), f"{output_path} not a directory?"

        self.output_path = output_path
        """destination directory"""
        self.emitters = {}
        """open file pointers"""
        self.verbose = verbose
        """log activity"""
        self.file_mode = file_mode
        """mode for file opens"""
        self.logger = logger

    def __enter__(self):
        """Ensure output_path exists, init emitter dict.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Close all open files."""
        for _ in self.emitters.values():
            _.close()
            if self.verbose:
                self.logger.info(f"closed {_.name}")

    def emit(self, name: str, file_mode: str = None) -> TextIO:
        """Maintain a hash of open files."""
        if not file_mode:
            file_mode = self.file_mode
        if name not in self.emitters:
            self.emitters[name] = open(self.output_path / f"{name}.ndjson", file_mode)
            if self.verbose:
                self.logger.info(f"opened {self.emitters[name].name}")

            # If the user already has metadata in the resource file but wants to add
            # another file the first line will create a 2nd record on the same line causing the commit step to error
            if file_mode != 'a':
                if os.path.getsize(self.emitters[name].name) > 0:
                    self.emitters[name].seek(self.emitters[name].tell() - 1, os.SEEK_SET)
                    last_char = self.emitters[name].read()
                    if last_char == "}":
                        self.emitters[name].write("\n")
        return self.emitters[name]


def validate_project_id(project_id) -> list[str]:
    """Ensure that the project_id is valid"""
    msgs = []
    if not project_id:
        msgs.append("project_id is missing")
    assert project_id.count('-') in [0, 1], f"{project_id} should have a single '-' delimiter."
    return msgs


def validate_email(email) -> list[str]:
    """Ensure that the email is valid"""
    msgs = []
    if not email:
        msgs.append("email is missing")
    if not email.count('@') == 1:
        msgs.append(f"{email} should have a single '@' delimiter.")
    try:
        from email_validator import validate_email as email_validator_validate, EmailNotValidError
        email_validator_validate(email)
    except EmailNotValidError as e:
        msgs.append(f"{email} is not a valid email address. {e}")
    return msgs


def to_resource_path(project_id):
    """Canonical conversion of project_id to resource path."""
    if '-' not in project_id:
        return project_id
    _ = project_id.split('-')
    return f"/programs/{_[0]}/projects/{_[1]}"


def unzip_collapse(zip_file, extract_to):
    """Unzip a file, collapse the directory structure."""
    c = 0
    with zipfile.ZipFile(zip_file, 'r') as zip_ref:
        for file_info in zip_ref.infolist():
            if file_info.is_dir():
                continue
            filename = os.path.basename(file_info.filename)
            extracted_path = os.path.join(extract_to, filename)
            with zip_ref.open(file_info.filename) as source, open(extracted_path, 'wb') as target:
                shutil.copyfileobj(source, target)
            c += 1
    assert c > 0, f"no files extracted from {zip_file}"


def calc_md5(file_path, md5_hash):
    """Calculate the md5 file, sort keys of json files."""
    read_binary = False
    with open(file_path, "r") as f:
        if file_path.name.endswith('ndjson'):
            for line in f.readlines():
                line = orjson.loads(line)
                md5_hash.update(orjson.dumps(line, option=orjson.OPT_SORT_KEYS))
        elif file_path.name.endswith('json'):
            data = orjson.load(f)
            md5_hash.update(orjson.dumps(data, option=orjson.OPT_SORT_KEYS))
        else:
            read_binary = True
    if read_binary:
        with open(file_path, "rb") as f:
            for line in f.readlines():
                md5_hash.update(line)
    return md5_hash


def identifier_to_string(identifier: list[Identifier]) -> str:
    """Return query parameter for identifier."""
    if identifier and not isinstance(identifier, list):
        identifier = [identifier]
    assert identifier and len(identifier) > 0, "identifier required"
    official = [_ for _ in identifier if _.use == 'official']
    if len(official) > 0:
        _ = official[0]
    else:
        _ = identifier[0]

    if _.system:
        return f"{_.system}|{_.value}"
    return _.value


def create_resource_id(resource, project_id) -> str:
    """Return id from identifier and project_id."""
    from gen3_tracker import ACED_NAMESPACE
    assert resource, "resource required"
    assert project_id, "project_id required"
    identifier_string = identifier_to_string(resource.identifier)
    return str(uuid.uuid5(ACED_NAMESPACE, f"{project_id}/{resource.resource_type}/{identifier_string}"))


def create_object_id(path: str, project_id: str) -> str:
    """Create a unique did for this object within a project i.e. DocumentReference."""
    from gen3_tracker import ACED_NAMESPACE

    def _normalize_file_url(_path: str) -> str:
        """Strip leading ./ and file:/// from file urls."""
        _path = re.sub(r'^file:\/\/\/', '', _path)
        _path = re.sub(r'^\.\/', '', _path)
        return _path

    base = f"{project_id}::{_normalize_file_url(path)}".encode()
    object_id = str(uuid.uuid5(ACED_NAMESPACE, base))

    return object_id


def assert_valid_id(resource, project_id):
    """Ensure that the id is correct."""
    assert resource, "resource required"
    assert project_id, "project_id required"
    if resource.resource_type == "DocumentReference":
        document_reference: DocumentReference = resource
        official_identifier = document_reference.content[0].attachment.url
        recreate_id = create_object_id(official_identifier, project_id)
    else:
        official = [_ for _ in resource.identifier if _.use == 'official']
        official_identifier = official[0].value if official else None
        recreate_id = create_resource_id(resource, project_id)
    if resource.id == recreate_id:
        return
    msg = f"The current {resource.resource_type}.id {resource.id} does not equal the calculated one {recreate_id}, has the project id changed? current:{project_id} {resource.resource_type}:{official_identifier}"
    raise Exception(msg)


class Commit(BaseModel):
    """A commit."""
    commit_id: str = None
    """The commit id."""
    object_id: str = None
    """The metadata file object_id."""
    message: str = None
    """The commit message."""
    resource_counts: dict = None
    """The resource counts of meta in this commit."""
    exceptions: list = None
    """The exceptions."""
    logs: list = None
    """The logs."""
    path: pathlib.Path = None
    """The path to the commit directory."""
    manifest_sqlite_path: pathlib.Path = None
    """The path to the manifest file."""
    meta_path: pathlib.Path = None
    """The path to the meta zip file."""


class Push(BaseModel):
    """A list of commits."""
    config: Config
    """The config."""

    commits: list[Commit] = []
    """A list of commits in this push."""

    published_timestamp: datetime = None
    """When the push was published."""

    published_job: dict = None

    def model_dump(self):
        """Dump the model.

         temporary until we switch to pydantic2
        """
        _ = self.model_dump_json(exclude={'config'})
        return json.loads(_)

    def pending_commits(self):
        """A list of commits yet to be pushed."""
        pending = []
        commits_dir = self.config.state_dir / self.config.gen3.project_id / 'commits'
        pending_path = commits_dir / 'pending.ndjson'
        if not pending_path.exists():
            return pending
        try:
            for _ in read_ndjson_file(pending_path):
                commit_dict = orjson.loads(open(commits_dir / _['commit_id'] / 'resource.json').read())
                commit = Commit(**commit_dict)

                commit.manifest_sqlite_path = commits_dir / _['commit_id'] / 'manifest.sqlite'
                commit.meta_path = commits_dir / _['commit_id'] / 'meta.zip'

                pending.append(commit)
            return pending
        except FileNotFoundError as e:
            print(f"No pending commits found in {pending_path} {e}")

    def pending_meta_index(self) -> list[dict]:
        """Index of pending meta files {id: resourceType}."""
        commits_dir = self.config.state_dir / self.config.gen3.project_id / 'commits'
        pending_path = commits_dir / 'pending.ndjson'
        pending = []
        if not pending_path.exists():
            return pending
        for _ in read_ndjson_file(pending_path):
            with open(commits_dir / _['commit_id'] / 'meta-index.ndjson') as fp:
                for line in fp.readlines():
                    pending.append(orjson.loads(line))
        return pending


def dict_md5(resource: dict) -> str:
    """Return the md5 of the dict."""
    return md5(orjson.dumps(resource, option=orjson.OPT_SORT_KEYS)).hexdigest()


def write_meta_index(index_path: pathlib.Path, source_path: pathlib.Path) -> pathlib.Path:
    """Write an index of ids and hashes from a path containing ndjson files.
    """
    index_path = index_path / 'meta-index.ndjson'
    with open(index_path, 'w') as fp:
        for _ in sorted(source_path.glob("*.ndjson")):
            for resource in read_ndjson_file(_):
                _ = {'id': resource['id'], 'resourceType': resource['resourceType'], 'md5': dict_md5(resource)}
                fp.write(orjson.dumps(_).decode())
                fp.write('\n')
    return index_path


def read_meta_index(index_path: pathlib.Path) -> Generator[dict, None, None]:
    """Read an index of ids and hashes from a path containing ndjson files.
    """
    index_path = index_path / 'meta-index.ndjson'
    if not index_path.exists():
        return
    with open(index_path, 'r') as fp:
        for line in fp.readlines():
            yield orjson.loads(line)


def to_metadata_dict(project_id=None, is_metadata=None, is_snapshot=None, md5=None, observation=None, patient=None, specimen=None, task=None):
    """Create metadata dict from parameters"""
    _ = {}
    if project_id:
        _['project_id'] = project_id
    if specimen:
        _['specimen_id'] = specimen
    if patient:
        _['patient_id'] = patient
    if task:
        _['task_id'] = task
    if observation:
        _['observation_id'] = observation
    if md5:
        _['md5'] = md5
    if is_metadata:
        _['is_metadata'] = is_metadata
    if is_snapshot:
        _['is_snapshot'] = is_snapshot
    return _


class CommandOutput(object):
    """Output object for commands."""

    def __init__(self):
        self.obj = None
        self.exit_code = 0

    def update(self, obj):
        """Update output with obj."""
        self.obj = obj


class CLIOutput:
    """Ensure output, exceptions and exit code are returned to user consistently."""
    from gen3_tracker import Config

    def __init__(self, config: Config, exit_on_error: bool = True):
        self.output = CommandOutput()
        self.config = config
        self.exit_on_error = exit_on_error

    def __enter__(self):
        return self.output

    def __exit__(self, exc_type, exc_val, exc_tb):
        rc = 0
        _ = {}
        if self.output.obj is not None:
            if isinstance(self.output.obj, dict):
                _.update(self.output.obj)
            elif isinstance(self.output.obj, list):
                _ = self.output.obj
            elif isinstance(self.output.obj, int):
                _ = {'count': self.output.obj}
            elif hasattr(self.output.obj, 'model_dump'):
                _.update(self.output.obj.model_dump())
            else:
                _.update(self.output.obj.dict())
        rc = self.output.exit_code
        if exc_type is not None:
            if isinstance(self.output.obj, dict):
                _['exception'] = f"{str(exc_val)}"
            elif isinstance(self.output.obj, list):
                _.append(f"{str(exc_val)}")
            else:
                _.update({'exception': f"{str(exc_val)}"})
            rc = 1
            logging.getLogger(__name__).exception(exc_val)
        if isinstance(_, dict) and 'msg' not in _:
            if rc == 1:
                _['msg'] = 'FAIL'
            else:
                _['msg'] = 'OK'
        prune = []
        if isinstance(_, dict):
            for k, v in _.items():
                if not v:
                    prune.append(k)
            for k in prune:
                del _[k]
        print_formatted(self.config, _)
        self.output.exit_code = rc
        if rc != 0 and self.exit_on_error:
            exit(rc)


def assert_config(config: Config):
    """Ensure that the config exists, if called without help"""
    try:
        if '--help' in sys.argv:
            return
        exists = (pathlib.Path('.g3t') / 'config.yaml').exists()
        msg = "project_id and profile are required"
        if not exists:
            msg = f"Config file config.yaml not found in {pathlib.Path('.g3t')}"
        assert config.gen3.project_id and config.gen3.profile, f"ERROR {msg}, have you run 'g3t init'?"
    except Exception as e:
        click.secho(str(e), fg='red')
        if config.debug:
            raise e
        sys.exit(1)


def parse_iso_tz_date(date_str: str) -> str:
    """Parse an iso date string."""
    # parse the string into a datetime object
    date_obj = dateutil_parser.parse(date_str)
    # if the date string doesn't have a timezone, you can add one
    if date_obj.tzinfo is None:
        date_obj = date_obj.replace(tzinfo=tzutc())
    # convert the datetime object back into an ISO formatted string with timezone
    return date_obj.isoformat()


def filter_dicts(data, pattern):
    """Filter a list of dictionaries based on a pattern."""
    pattern = re.compile(pattern)

    if not isinstance(data, list):
        data = [data]

    def match_pattern(value):
        if isinstance(value, str):
            return pattern.search(value) is not None
        elif isinstance(value, dict):
            return any(match_pattern(v) for v in value.values())
        elif isinstance(value, list):
            return any(match_pattern(v) for v in value)
        else:
            return False

    return [d for d in data if any(match_pattern(v) for v in d.values())]



import importlib
import logging
import pathlib
from collections import defaultdict
from typing import Iterator, Any, Optional, Callable

# TODO fix me, make configurable
from fhir.resources import FHIRAbstractModel
from nested_lookup import nested_lookup
from pydantic.v1 import ValidationError, validator
from pydantic import BaseModel, ConfigDict

from gen3_tracker.common import is_json_extension, read_json, read_ndjson_file

FHIR_CLASSES = importlib.import_module('fhir.resources')

logger = logging.getLogger(__name__)


class ParseResult(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    """Results of FHIR validation of dict."""
    resource: Optional[Any] = None
    """If valid, the FHIR resource."""
    exception: Optional[Exception] = None
    """If invalid, the exception."""
    path: Optional[pathlib.Path] = None
    """Source file, if available."""
    offset: Optional[int] = 0
    """Base 0 offset of line number(ndjson) or entry(bundle)."""
    resource_id: Optional[str] = None
    """Resource id of resource"""
    json_obj: Optional[dict] = None
    """Original json object"""

    @validator("resource")
    def validate_resource(cls, val):
        if val is None:
            return val
        if issubclass(type(val), FHIRAbstractModel):
            return val
        raise TypeError(f"Wrong type for 'resource', was {type(val)} must be subclass of FHIRAbstractModel")


def parse_obj(resource: dict, validate=True) -> ParseResult:
    """Load a dictionary into a FHIR model """
    try:
        assert 'resourceType' in resource, "Dict missing `resourceType`, is it a FHIR dict?"
        klass = FHIR_CLASSES.get_fhir_model_class(resource['resourceType'])
        _ = klass.parse_obj(resource)
        if validate:
            # trigger object traversal, see monkey patch below, at bottom of file
            _.dict()
        return ParseResult(resource=_, exception=None, path=None, resource_id=_.id)
    except (ValidationError, AssertionError) as e:
        return ParseResult(resource=None, exception=e, path=None, resource_id=resource.get('id', None))


def _entry_iterator(parse_result: ParseResult) -> Iterator[ParseResult]:
    """See if there are entries"""
    if not _has_entries(parse_result):
        yield parse_result
    else:
        _path = parse_result.path
        offset = 0
        if parse_result.resource.entry and len(parse_result.resource.entry) > 0:
            for _ in parse_result.resource.entry:
                if _ is None:
                    break
                if hasattr(_, 'resource') and _.resource:  # BundleEntry
                    yield ParseResult(path=_path, resource=_.resource, offset=offset, exception=None, json_obj=_.resource.dict())
                elif hasattr(_, 'item'):  # ListEntry
                    yield ParseResult(path=_path, resource=_.item, offset=offset, exception=None, json_obj=_.item.dict())
                else:
                    yield ParseResult(path=_path, resource=_.item, offset=offset, exception=None, json_obj=_.item.dict())
                offset += 1
    pass


def _has_entries(_: ParseResult):
    """FHIR types Bundles List have entries"""
    if _.resource is None:
        return False
    return _.resource.resource_type in ["List"] and _.resource.entry is not None


def directory_reader(directory_path: str,
                     recurse: bool = True,
                     validate: bool = False) -> Iterator[ParseResult]:
    """Extract FHIR resources from directory

    Read any type of json file, return itemized resources by iterating through Bundles and Lists
    """

    if isinstance(directory_path, str):
        directory_path = pathlib.Path(directory_path)

    directory_path = directory_path.expanduser()

    try:
        input_files = [_ for _ in pathlib.Path.glob(directory_path.name) if is_json_extension(_.name)]
    except TypeError:
        input_files = []

    if len(input_files) == 0:
        if recurse:
            input_files = [_ for _ in directory_path.glob('**/*.*') if is_json_extension(_.name)]

    # assert len(input_files) > 0, f"No files found in {directory_path.name}"

    for input_file in input_files:
        offset = 0
        for json_obj in read_json(input_file):
            parse_result = parse_obj(json_obj, validate=validate)
            parse_result.path = input_file
            parse_result.offset = offset
            parse_result.json_obj = json_obj
            offset += 1
            for _ in _entry_iterator(parse_result):
                yield _


def aggregate(metadata_path: pathlib.Path | str) -> dict:
    """Aggregate metadata counts resourceType(count)-count->resourceType(count)."""

    nested_dict: Callable[[], defaultdict[str, defaultdict]] = lambda: defaultdict(defaultdict)

    if not isinstance(metadata_path, pathlib.Path):
        metadata_path = pathlib.Path(metadata_path)
    summary = nested_dict()
    for path in sorted(metadata_path.glob("*.ndjson")):
        for _ in read_ndjson_file(path):

            resource_type = _['resourceType']
            if 'count' not in summary[resource_type]:
                summary[resource_type]['count'] = 0
            summary[resource_type]['count'] += 1

            refs = nested_lookup('reference', _)
            for ref in refs:
                # A codeable reference is an object with a codeable concept and a reference
                if isinstance(ref, dict):
                    ref = ref['reference']
                ref_resource_type = ref.split('/')[0]
                if 'references' not in summary[resource_type]:
                    summary[resource_type]['references'] = nested_dict()
                dst = summary[resource_type]['references'][ref_resource_type]
                if 'count' not in dst:
                    dst['count'] = 0
                dst['count'] += 1

    return summary



import logging
from typing import List

import yaml
from gen3.auth import Gen3Auth
from pydantic import BaseModel

from gen3_tracker.collaborator.access import get_requests, get_request, create_request, update_request
from gen3_tracker.config import Config, ensure_auth

from importlib.resources import files, as_file
from . import policies


class LogAccess(BaseModel):
    endpoint: str = None
    """The commons url"""
    requests: List[dict] = None
    """List of requests"""
    request: dict = None
    """A single request"""
    msg: str = None
    """A message"""
    commands: List[str] = None
    """A list of commands to run"""


def format_policy(policy: dict, project_id: str, user_name: str) -> dict:
    """Format policy by applying project_id and user_name.

    Parameters:
    policy (dict): policy to format
    project_id (str): project_id to apply to policy's resource_path PROGRAM PROJECT tokens
    user_name (str): user_name to apply to policy's username
    """
    if user_name and policy.get('username', None) is None:
        policy['username'] = user_name
    if project_id:
        program, project = project_id.split('-')
        if policy.get('resource_paths', None):
            policy['resource_paths'] = [_.replace('PROGRAM', program).replace('PROJECT', project) for _ in policy['resource_paths']]
        elif policy.get('policy_id', None):
            policy['policy_id'] = policy['policy_id'].replace('PROGRAM', program).replace('PROJECT', project)
        else:
            raise ValueError(f"No resource_paths or policy_id specified, can't apply project_id {policy}")
        policy['resource_display_name'] = f"{project_id}"
    else:
        if 'PROGRAM' in policy['resource_path'] or 'PROJECT' in policy['resource_path']:
            raise ValueError(f"specify project_id for {policy['resource_path']}")
    return policy


def ls(config: Config, mine: bool, active: bool = False, username: str = None, auth: Gen3Auth = None) -> LogAccess:
    """List requests."""
    if not auth:
        auth = ensure_auth(config=config)
    assert auth, "auth required"
    requests = get_requests(auth=auth, mine=mine, active=active, username=username)
    if not isinstance(requests, list):
        raise Exception(f"Unexpected response: {requests}")
    return LogAccess(**{
        'endpoint': auth.endpoint,
        'requests': [_ for _ in requests],
    })


def cat(config: Config, request_id: str) -> dict:
    """Show a specific request requests."""
    auth = ensure_auth(config=config)
    request = get_request(auth=auth, request_id=request_id)
    return LogAccess(**{
        'endpoint': auth.endpoint,
        'requests': [request],
    })


def cp(config: Config, request: dict, revoke: bool = False, auth: Gen3Auth = None) -> LogAccess:
    """List requests."""

    if not auth:
        auth = ensure_auth(config=config)

    request = create_request(auth=auth, request=request, revoke=revoke)
    return LogAccess(**{
        'endpoint': auth.endpoint,
        'request': request,
    })


ALLOWED_REQUEST_STATUSES = """DRAFT SUBMITTED APPROVED SIGNED REJECTED""".split()


def update(config: Config, request_id: str, status: str, auth: Gen3Auth = None) -> LogAccess:
    """Update request."""
    assert request_id, "required"
    assert status, "required"
    status = status.upper()
    assert status in ALLOWED_REQUEST_STATUSES, f"{status} not in {ALLOWED_REQUEST_STATUSES}"

    if not auth:
        auth = ensure_auth(config=config)

    request = update_request(auth=auth, request_id=request_id, status=status)
    return LogAccess(**{
        'endpoint': auth.endpoint,
        'request': request,
    })


def add_user(config: Config, project_id: str, user_name: str, write: bool, delete: bool, auth: Gen3Auth, existing_requests=[]) -> LogAccess:
    """Add user to project by assigning them policies."""

    # implement read from resource_path
    policies_ = []
    file_names = ['add-user-read.yaml']
    if write:
        file_names.append('add-user-write.yaml')
    if delete:
        file_names.append('add-user-delete.yaml')

    for file_name in file_names:
        source = files(policies).joinpath(file_name)
        with as_file(source) as file_path:
            with open(file_path) as f:
                policies_.extend([_ for _ in yaml.safe_load(f)['policies']])

    requests = []
    request_ids = []
    for policy in policies_:
        policy = format_policy(policy, project_id, user_name)
        if policy['policy_id'] in [_['policy_id'] for _ in existing_requests]:
            print("DBG policy already exists", policy['policy_id'])
            continue

        requests.append(cp(request=policy, config=config, auth=auth).request)
        request_ids.append(requests[-1]['request_id'])

    commands = ["g3t collaborator add <user> --approve"]
    msg = f"An authorized user must approve these requests to  add {user_name} to {project_id}"

    return LogAccess(**{
        'requests': requests,
        'commands': commands,
        'msg': msg,
    })


def rm_user(config: Config, project_id: str, user_name: str) -> LogAccess:
    """Revoke user from project's policies."""

    # implement read from resource_path
    policies_ = []
    file_names = ['add-user-read.yaml', 'add-user-write.yaml']

    for file_name in file_names:
        source = files(policies).joinpath(file_name)
        with as_file(source) as file_path:
            with open(file_path) as f:
                policies_.extend([_ for _ in yaml.safe_load(f)['policies']])

    requests = []
    request_ids = []
    for policy in policies_:
        policy = format_policy(policy, project_id, user_name)
        try:
            requests.append(cp(request=policy, config=config, revoke=True).request)
        except Exception as e:
            logging.getLogger(__package__).warning(f"Failed to revoke {user_name} from {project_id}: {e}")
    commands = [f"g3t utilities access update {request_id} SIGNED" for request_id in request_ids]
    msg = f"Approve these requests to rm {user_name} to {project_id}"

    return LogAccess(**{
        'requests': requests,
        'commands': commands,
        'msg': msg,
    })


def add_policies(config: Config, project_id: str, auth: Gen3Auth = None) -> LogAccess:
    """Add policies to project. """
    # implement read from resource_path
    policies_ = []
    file_names = ['add-project-default.yaml']

    for file_name in file_names:
        source = files(policies).joinpath(file_name)
        with as_file(source) as file_path:
            with open(file_path) as f:
                policies_.extend([_ for _ in yaml.safe_load(f)['policies']])

    requests = []
    request_ids = []
    for policy in policies_:
        policy = format_policy(policy, project_id, user_name=None)
        requests.append(cp(request=policy, config=config, auth=auth).request)
        request_ids.append(requests[-1]['request_id'])

    commands = []
    msg = "OK"
    return LogAccess(**{
        'requests': requests,
        'msg': msg,
        'commands': commands,
    })



import sys

import click
from halo import Halo

from gen3_tracker import NaturalOrderGroup
from gen3_tracker.collaborator.access.requestor import update
from gen3_tracker.common import CLIOutput, assert_config, ERROR_COLOR, validate_email
from gen3_tracker.config import Config, ensure_auth


@click.group(name='collaborator', cls=NaturalOrderGroup)
@click.pass_context
def collaborator(ctx):
    """Manage project membership."""
    cmd = ctx.invoked_subcommand
    if cmd not in ['add-steward', 'pending', 'approve', 'add']:
        config: Config = ctx.obj
        assert_config(config)


@collaborator.command(name="add")
@click.argument('username', required=True, type=str)
@click.argument('resource_path', required=False, type=str)
@click.option('--write/--no-write', '-w', help='Give user write privileges', is_flag=True, default=False, show_default=True)
# @click.option('--delete/--no-delete', '-d', help='Give user delete privileges', is_flag=True, default=False, show_default=True)
@click.option('--approve', '-a', help='Approve the addition (privileged)', is_flag=True, default=False, show_default=True)
@click.pass_obj
def project_add_user(config: Config, username: str, resource_path: str, write: bool, approve: bool):
    """Add user to project."""
    import gen3_tracker.collaborator.access.requestor
    from gen3_tracker.collaborator.access.requestor import add_user, update

    assert username, "username (email) required"
    if resource_path:
        resource_path = resource_path.split('/')
        program = resource_path[2]
        project = resource_path[4]
        project_id = f"{program}-{project}"
    else:
        project_id = config.gen3.project_id
        program = config.gen3.program
        project = config.gen3.project

    assert program, "program required"
    assert project, "project required"

    with CLIOutput(config=config) as output:
        try:
            with Halo(text='Searching', spinner='line', placement='right', color='white'):
                auth = ensure_auth(config=config)
                existing_requests = gen3_tracker.collaborator.access.requestor.ls(config=config, mine=False, auth=auth, username=username).requests
                existing_requests = [r for r in existing_requests if r['resource_display_name'] == project_id]
                needs_approval = []

                _ = add_user(config, project_id, username, write, delete=False, auth=auth, existing_requests=existing_requests)
                existing_requests = _.requests

                for request in existing_requests:
                    if request['status'] != 'SIGNED':
                        needs_approval.append(request)

            if approve and not needs_approval:
                click.secho(f"User {username} already has approved requests for {project_id}.", fg='yellow')
                output.update({'existing': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in existing_requests]})
            elif not approve and needs_approval:
                output.update(
                    {
                        'needs_approval': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in needs_approval],
                        'msg': f"An authorized user must approve these requests to add {username} to {project_id} see --approve",
                    })
            else:
                approvals = []
                with Halo(text='Approving', spinner='line', placement='right', color='white'):
                    for request in needs_approval:
                        approvals.append(update(config, request_id=request['request_id'], status='SIGNED', auth=auth).request)
                output.update({'approved': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in approvals]})

        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e


@collaborator.command(name="rm")
@click.argument('username', required=True, type=str)
@click.option('--approve', '-a', help='Approve the removal (privileged)', is_flag=True, default=False, show_default=True)
@click.pass_obj
def project_rm_user(config: Config, username: str, approve: bool):
    """Remove user from project."""
    from gen3_tracker.collaborator.access.requestor import rm_user, update

    with CLIOutput(config=config) as output:
        try:
            assert username, "username (email) required"
            project_id = config.gen3.project_id

            approvals = []
            needs_approval = []
            auth = ensure_auth(config=config)
            with Halo(text='Removing', spinner='line', placement='right', color='white'):
                _ = rm_user(config, project_id, username)
                needs_approval.extend(_.requests)
            if approve:
                with Halo(text='Approving', spinner='line', placement='right', color='white'):
                    for request in needs_approval:
                        _ = update(config, request_id=request['request_id'], status='SIGNED', auth=auth).request
                        approvals.append(_)
                output.update({'approved': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status']} for r in approvals]})
            else:
                output.update({'needs_approval': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status']} for r in needs_approval]})

        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e


@collaborator.command(name="ls")
@click.pass_obj
def project_ls_user(config: Config):
    """List all requests in project."""
    import gen3_tracker.collaborator.access.requestor

    with CLIOutput(config=config) as output:
        try:
            program = config.gen3.program
            project = config.gen3.project
            project_id = f"{program}-{project}"

            with Halo(text='Searching', spinner='line', placement='right', color='white'):
                auth = ensure_auth(config=config)
                # get all requests
                existing_requests = gen3_tracker.collaborator.access.requestor.ls(config=config, mine=False, auth=auth).requests
                # filter for project
                existing_requests_for_project = [r for r in existing_requests if r.get('resource_display_name', None) == project_id]
                output.update({
                                'existing': [{'policy_id': r['policy_id'],
                                              'resource_display_name': r['resource_display_name'],
                                              'request_id': r['request_id'],
                                              'status': r['status'],
                                              'username': r['username'],
                                              'updated_time': r['updated_time']
                                              } for r in existing_requests_for_project]})

        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e


@collaborator.command(name="pending")
@click.pass_obj
def project_ls_pending(config: Config):
    """Show all pending requests."""
    import gen3_tracker.collaborator.access.requestor

    with CLIOutput(config=config) as output:
        try:
            with Halo(text='Searching', spinner='line', placement='right', color='white'):
                auth = ensure_auth(config=config)
                existing_requests = gen3_tracker.collaborator.access.requestor.ls(config=config, mine=False,
                                                                                  auth=auth).requests
                needs_approval = []
                for request in existing_requests:
                    if request['status'] != 'SIGNED':
                        needs_approval.append(request)

                output.update(
                    {'existing': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in needs_approval]}
                )

        except Exception as e:
            click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
            output.exit_code = 1
            if config.debug:
                raise e


@collaborator.command(name="approve")
@click.option('--request_id', required=False, help='Sign only this request')
@click.option('--all', 'all_requests', required=False, is_flag=True, help='Sign all requests')
@click.pass_obj
def project_approve_request(config: Config, request_id: str, all_requests: bool):
    """Sign an existing request (privileged)."""
    import gen3_tracker.collaborator.access.requestor
    from gen3_tracker.collaborator.access.requestor import update

    with CLIOutput(config=config) as output:
        try:
            assert request_id or all_requests, "request_id or --all required"
            with Halo(text='Signing', spinner='line', placement='right', color='white'):
                auth = ensure_auth(config=config)
                program = config.gen3.program
                project = config.gen3.project
                if all_requests:
                    existing_requests = gen3_tracker.collaborator.access.requestor.ls(config=config, mine=False,
                                                                                      auth=auth).requests

                    if program and project:
                        existing_requests = [r for r in existing_requests if r['resource_display_name'] == f"{program}-{project}"]
                    needs_approval = []
                    for request in existing_requests:
                        if request['status'] != 'SIGNED':
                            needs_approval.append(request)

                    approved = []
                    assert len(needs_approval) > 0, "No requests to approve.  You must be a privileged user to approve requests."
                    for request in needs_approval:
                        approved.append(update(config, request_id=request['request_id'], status='SIGNED', auth=auth).request)
                    output.update(
                        {'approved': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in approved]}
                    )
                else:
                    r = update(config, request_id=request_id, status='SIGNED', auth=auth).request
                    output.update(
                        {'approved': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']}]}
                    )

        except Exception as e:
            click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
            output.exit_code = 1
            if config.debug:
                raise e


@collaborator.command(name="add-steward", hidden=True)
@click.argument('user_name', required=True)
@click.argument('resource_path', required=True)
@click.option('--approve', '-a', help='Approve the addition (privileged)', is_flag=True, default=False, show_default=True)
@click.pass_obj
def add_steward(config: Config, resource_path: str, user_name: str, approve: bool):
    """Add a data steward user with approval rights to a program.

    \b
    USER_NAME (str): user's email
    RESOURCE_PATH (str): Gen3 authz /programs/<program>
    """
    from gen3_tracker.collaborator.access import create_request

    with CLIOutput(config=config) as output:
        try:

            msgs = validate_email(user_name)
            assert msgs == [], f"Invalid email address: {user_name} {msgs}"
            assert user_name, "user_name required"

            request = {"username": user_name, "resource_path": resource_path}
            roles = ['requestor_reader_role', 'requestor_updater_role']

            needs_approval = []
            approvals = []
            with Halo(text='Adding', spinner='line', placement='right', color='white'):
                auth = ensure_auth(config=config)

                user = auth.curl('/user/user').json()
                is_privileged = False
                for _ in user['authz']['/programs']:
                    if _['method'] == 'update' and _['service'] == 'requestor':
                        is_privileged = True
                        break
                assert is_privileged, "You must be a privileged user to add a steward."

                for role in roles:
                    request.update({"role_ids": [role]})
                    needs_approval.append(create_request(config=config, request=request))

            if approve:
                with Halo(text='Approving', spinner='line', placement='right', color='white'):
                    for request in needs_approval:
                        approvals.append(update(config, request_id=request['request_id'], status='SIGNED', auth=auth).request)
                output.update({'approved': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in approvals]})
            else:
                output.update({'needs_approval': [{'policy_id': r['policy_id'], 'request_id': r['request_id'], 'status': r['status'], 'username': r['username']} for r in needs_approval]})

        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1
            if config.debug:
                raise e



import pathlib
from collections import Counter
from collections import defaultdict
from typing import List
from urllib.parse import urlparse

import orjson
from fhir.resources.coding import Coding
from fhir.resources.identifier import Identifier
from fhir.resources.reference import Reference
from nested_lookup import nested_lookup
from pydantic import BaseModel, ConfigDict

from gen3_tracker.common import assert_valid_id
from gen3_tracker.meta import ParseResult, directory_reader


class ValidateDirectoryResult(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    """Results of FHIR validation of directory."""
    resources: dict
    exceptions: List[ParseResult]

    def model_dump(self):
        """

         temporary until we switch to pydantic2
        """
        for _ in self.exceptions:
            _.exception = str(_.exception)
            _.path = str(_.path)
        return orjson.loads(self.model_dump_json())


def _check_coding(self: Coding, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Coding."""
    # note `self` is the Coding
    assert self.code, f"Missing `code` {self}"
    assert (not self.code.startswith("http")), f"`code` should _not_ be a url http {self.code}"
    assert ":" not in self.code, f"`code` should not contain ':' {self.code}"
    assert self.system, f"Missing `system` {self}"
    assert "%" not in self.system, f"`system` should be a simple url without uuencoding {self.system}"
    parsed = urlparse(self.system)
    assert parsed.scheme, f"`system` is not a URI {self}"
    assert self.display, f"Missing `display` {self}"
    # call the original dict() method
    return orig_coding_dict(self, *args, **kwargs)


def _check_identifier(self: Identifier, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Identifier."""
    # note `self` is the Identifier
    assert self.value, f"Missing `value` {self}"
    assert self.system, f"Missing `system` {self}"
    parsed = urlparse(self.system)
    assert parsed.scheme, f"`system` is not a URI {self}"
    assert "%" not in self.system, f"`system` should be a simple url without uuencoding {self.system}"
    # call the original dict() method
    return orig_identifier_dict(self, *args, **kwargs)


def _check_reference(self: Reference, *args, **kwargs):
    """MonkeyPatch replacement for dict(), check Reference."""
    # note `self` is the Identifier
    assert self.reference, f"Missing `reference` {self}"
    assert '/' in self.reference, f"Does not appear to be Relative reference {self}"
    assert 'http' not in self.reference, f"Absolute references not supported {self}"
    assert len(self.reference.split('/')) == 2, f"Does not appear to be Relative reference {self}"

    # call the original dict() method
    return orig_reference_dict(self, *args, **kwargs)


def validate(directory_path: pathlib.Path, project_id=None) -> ValidateDirectoryResult:
    """Check FHIR data, accumulate results."""
    exceptions = []
    resources = defaultdict(int)
    # add resources to bundle
    references = []
    ids = []

    for parse_result in directory_reader(directory_path, validate=True):
        if parse_result.exception:
            exceptions.append(parse_result)
            continue

        try:
            assert_valid_id(parse_result.resource, project_id)
        except Exception as e:
            parse_result.exception = e
            exceptions.append(parse_result)
            continue

        _ = parse_result.resource
        ids.append(f"{_.resource_type}/{_.id}")
        nested_references = nested_lookup('reference', parse_result.json_obj)
        # https://www.hl7.org/fhir/medicationrequest-definitions.html#MedicationRequest.medication
        # is a reference to a Medication resource https://www.hl7.org/fhir/references.html#CodeableReference
        # so it has a reference.reference form, strip it out
        nested_references = [_ for _ in nested_references if isinstance(_, str)]
        references.extend(nested_references)
        resources[parse_result.resource.resource_type] += 1

    # assert references exist
    references = set(references)
    ids_list = ids
    ids = set(ids)
    if not references.issubset(ids):
        _ = Exception(f"references not found {references - ids}")
        _ = ParseResult(resource=None, exception=_, path=directory_path, resource_id=None)
        exceptions.append(_)
    if len(ids) != len(ids_list):
        # Create a Counter object from ids_list
        counter = Counter(ids_list)
        # Get the duplicate ids
        duplicate_ids = [id_ for id_, count in counter.items() if count > 1]
        # log it
        _ = Exception(f"Duplicate ids found {duplicate_ids}")
        _ = ParseResult(resource=None, exception=_, path=directory_path, resource_id=None)
        exceptions.append(_)

    return ValidateDirectoryResult(resources={'summary': dict(resources)}, exceptions=exceptions)


#
# monkey patch dict() methods
#
orig_coding_dict = Coding.dict
Coding.dict = _check_coding

orig_identifier_dict = Identifier.dict
Identifier.dict = _check_identifier

orig_reference_dict = Reference.dict
Reference.dict = _check_reference



from gen3.auth import Gen3Auth

from gen3_tracker.config import Config, ensure_auth
from gen3_tracker.projects import ProjectSummaries, get_projects, ProjectSummary


def ls(config: Config, resource_filter: str = None, msgs: list[str] = [], auth: Gen3Auth = None, full: bool = True) -> ProjectSummaries:
    """List projects."""
    # improve startup time by importing only what is needed

    if not auth:
        auth = ensure_auth(config=config)

    projects = get_projects(auth)

    if full:
        project_messages = {'complete': {}, 'incomplete': {}}
        for _program in projects:
            for _project in projects[_program]:
                if resource_filter and resource_filter != f"/programs/{_program}/projects/{_project}":
                    continue
                _ = 'complete'
                if not projects[_program][_project]['exists']:
                    _ = 'incomplete'
                project_messages[_][
                    f"/programs/{_program}/projects/{_project}"
                ] = ProjectSummary(
                    user_perms=True if _ == 'complete' else False,
                    permissions=projects[_program][_project]['permissions'],
                )
    else:
        project_messages = {'complete': [], 'incomplete': []}
        any_incomplete = False
        for _program in projects:
            for _project in projects[_program]:
                if resource_filter and resource_filter != f"/programs/{_program}/projects/{_project}":
                    continue
                if not projects[_program][_project]['exists']:
                    any_incomplete = True
                    project_messages['incomplete'].append(f"/programs/{_program}/projects/{_project}")
                else:
                    project_messages['complete'].append(f"/programs/{_program}/projects/{_project}")
        if any_incomplete:
            msgs.append("some available projects are missing read/write permissions")
        else:
            msgs.append("user has read/write permissions to all available projects")

    if len(project_messages) == 0:
        msgs.append("No projects found.")

    return ProjectSummaries(**{
        'endpoint': auth.endpoint,
        'incomplete': project_messages['incomplete'],
        'complete': project_messages['complete'],
        'messages': msgs
    })



###########################
# LOCAL FHIR DATABASE ###
###########################

import inflection
import json
import ndjson
import numpy as np
import pandas as pd
import pathlib
import sqlite3

from collections import defaultdict
from deepmerge import always_merger
from functools import lru_cache
from typing import Dict, Generator, List

from gen3_tracker.meta.entities import (
    SimplifiedResource,
    get_nested_value,
    normalize_coding,
    normalize_value,
    traverse,
)


class LocalFHIRDatabase:
    def __init__(self, db_name):  # , db_name=pathlib.Path('.g3t') / 'local.db'):
        self.db_name = db_name
        self.connection = None
        self.cursor = None
        self.table_created = {}  # Flag to track if the table has been created

    def connect(self) -> sqlite3.Cursor:
        """establish database connection if not established, return cursor"""
        if self.connection is None:
            self.connection = sqlite3.connect(self.db_name)
        if self.cursor is None:
            self.cursor = self.connection.cursor()
        else:
            return self.cursor

    def disconnect(self) -> None:
        """clean up database connection"""
        if self.connection:
            self.connection.commit()
            self.connection.close()

    def create_table(
        self,
        name="resources",
        ddl="""
                    CREATE TABLE __NAME__ (
                        key TEXT PRIMARY KEY,
                        resource_type TEXT,
                        resource JSON
                    )
                """,
    ):
        self.connect()
        # Check if the table exists before creating it
        self.cursor.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{name}'"
        )
        table_exists = self.cursor.fetchone()

        if not table_exists:
            ddl = ddl.replace("__NAME__", name)
            self.cursor.execute(ddl)
            self.table_created[name] = True

    def count(self, table_name="resources"):
        self.connect()
        self.cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
        count = self.cursor.fetchone()[0]
        return count

    def insert_data(self, id_, resource_type, resource, table_name="resources"):
        """Insert data into the database."""
        if table_name not in self.table_created:
            self.create_table(
                table_name
            )  # Lazily create the table if not already created

        composite_key = f"{resource_type}/{id_}"

        # see if the resource already exists
        self.cursor.execute(
            """
            SELECT resource FROM resources WHERE key = ?
        """,
            (composite_key,),
        )
        row = self.cursor.fetchone()

        # Initialize an empty dictionary to hold the resource
        existing_resource = {}

        # Check if the row is not None
        if row is not None:
            # The first element of the row contains the resource as a JSON string
            resource_json = row[0]

            # Convert the JSON string into a Python dictionary
            existing_resource = json.loads(resource_json)

        # Merge the existing resource with the new resource
        resource = always_merger.merge(existing_resource, resource)

        self.cursor.execute(
            f"""
            INSERT INTO {table_name} (key, resource_type, resource)
            VALUES (?, ?, ?)
        """,
            (composite_key, resource_type, json.dumps(resource)),
        )
        # print(f"Inserted {composite_key} into the database")

    def insert_data_from_dict(self, resource, table_name="resources"):
        """Insert data into the database from a dictionary."""
        if "id" not in resource or (
            "resource_type" not in resource and "resourceType" not in resource
        ):
            raise ValueError(
                f"Resource dictionary must contain 'id' and 'resource_type' keys {resource}"
            )
        self.insert_data(
            resource["id"],
            resource.get("resource_type", resource.get("resourceType")),
            resource,
            table_name,
        )

    def bulk_insert_data(self, resources, table_name="resources") -> int:
        """Bulk insert data into the database."""

        if table_name not in self.table_created:
            self.create_table(
                table_name
            )  # Lazily create the table if not already created

        def _prepare(resource):
            resource_type = resource.get("resource_type", resource.get("resourceType"))
            id_ = resource["id"]
            composite_key = f"{resource_type}/{id_}"
            return (composite_key, resource_type, json.dumps(resource))

        def _iterate(_resources):
            for _ in _resources:
                yield _prepare(_)

        try:
            self.connect()
            sql = f"""
                INSERT INTO {table_name} (key, resource_type, resource)
                VALUES (?, ?, ?)
            """
            new_cursor = self.cursor.executemany(sql, _iterate(_resources=resources))

        except sqlite3.IntegrityError as e:
            for resource in resources:
                prepared_resource = _prepare(resource)
                try:
                    self.cursor.execute(sql, prepared_resource)
                except sqlite3.IntegrityError:
                    print(f"Error inserting resource: {prepared_resource}")
                    print(f"Exception: {e}")
            raise

        finally:
            self.connection.commit()
            # self.disconnect()

        return new_cursor.rowcount

    def load_from_ndjson_file(self, file_path, table_name="resources"):
        """Load the NDJSON file into the database."""

        if table_name not in self.table_created:
            self.create_table(
                table_name
            )  # Lazily create the table if not already created

        with open(file_path, "r") as file:
            reader = ndjson.reader(file)
            self.bulk_insert_data(reader)

    def load_ndjson_from_dir(self, path: str = "META", pattern: str = "*.ndjson"):
        """Load all the NDJSON files in the directory into the database."""
        for file_path in pathlib.Path(path).glob(pattern):
            self.load_from_ndjson_file(file_path)

    @lru_cache(maxsize=None)
    def patient_everything(self, patient_id) -> Generator[dict, None, None]:
        """Return all the resources for a patient."""
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT * FROM resources WHERE key = ?", (f"Patient/{patient_id}",)
        )

        for _ in cursor.fetchall():
            key, resource_type, resource = _
            yield json.loads(resource)

    # @lru_cache(maxsize=None) this cache was giving updated lookup results when I di
    def patient(self, patient_id) -> dict:
        """Return the patient resource."""
        cursor = self.cursor
        self.cursor.execute(
            "SELECT * FROM resources WHERE key = ?", (f"Patient/{patient_id}",)
        )
        _ = cursor.fetchone()
        if _ is None:
            print(f"Patient {patient_id} not found")
            return None
        key, resource_type, resource = _
        resource = json.loads(resource)

        resource = self.simplify_extensions(resource)

        return resource

    @lru_cache(maxsize=None)
    def condition_everything(self) -> List[Dict]:
        """Return all the resources for a Condition."""
        cursor = self.connection.cursor()
        cursor.execute(
            " SELECT * FROM resources where resource_type = ?", ("Condition",)
        )

        resources = []
        for row in cursor.fetchall():
            key, resource_type, resource = row
            resources.append(json.loads(resource))

        return resources

    @lru_cache(maxsize=None)
    def resource(self, resourceType, id) -> dict:
        """Return any resource with id and type specified"""
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT * FROM resources WHERE key = ?", (f"{resourceType}/{id}",)
        )
        _ = cursor.fetchone()
        if _ is None:
            print(f"{resourceType} {id} not found")
            return None
        key, resource_type, resource = _
        resource = json.loads(resource)

        resource = self.simplify_extensions(resource)

    @staticmethod
    def simplify_extensions(resource: dict) -> dict:
        """Extract extension values, derive key from extension url"""
        for _ in resource.get("extension", []):
            value_normalized, value_source = normalize_value(_)
            extension_key = _["url"].split("/")[-1]
            extension_key = (
                inflection.underscore(extension_key)
                .removesuffix(".json")
                .removeprefix("structure_definition_")
            )
            resource[extension_key] = value_normalized
            assert value_normalized, f"extension: {extension_key} = {value_normalized}"
        if "extension" in resource:
            del resource["extension"]
        return resource

    # @lru_cache(maxsize=None)
    def flattened_procedure(self, procedure_key) -> dict:
        """Return the procedure with everything resolved."""
        cursor = self.cursor
        cursor.execute("SELECT * FROM resources WHERE key = ?", (procedure_key,))
        key, resource_type, resource = cursor.fetchone()
        procedure = json.loads(resource)

        # simplify the identifier
        procedure["identifier"] = procedure["identifier"][0]["value"]
        # simplify the code
        procedure["code"] = procedure["code"]["coding"][0]["display"]
        # simplify the reason
        procedure["reason"] = procedure["reason"][0]["reference"]["reference"]
        # simplify the occurrenceAge
        procedure["occurrenceAge"] = procedure["occurrenceAge"]["value"]
        # simplify the subject
        subject = procedure["subject"]["reference"]
        procedure["subject"] = subject

        return procedure

    @lru_cache(maxsize=None)
    def flattened_condition(self, condition_key) -> dict:
        """Return the procedure with everything resolved."""
        #  TODO - implement with gen3_tracker.meta.entites.SimplifiedFHIR
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (condition_key,))
        key, resource_type, resource = cursor.fetchone()
        condition = json.loads(resource)

        # simplify the identifier
        condition["identifier"] = get_nested_value(
            condition, ["identifier", 0, "value"]
        )

        # simplify the code
        condition["code"] = self.select_coding(condition)
        condition["category"] = self.select_category(condition)

        for coding_normalized, coding_source in normalize_coding(condition):
            condition[coding_source] = coding_normalized
        # simplify the onsetAge
        condition["onsetAge"] = get_nested_value(condition, ["onsetAge", "value"])

        return condition

    def flattened_procedures(self) -> Generator[dict, None, None]:
        #  TODO - implement with gen3_tracker.meta.entites.SimplifiedFHIR
        """Return all the procedures with everything resolved"""
        loaded_db = self
        connection = sqlite3.connect(loaded_db.db_name)
        cursor = connection.cursor()
        cursor.execute(
            "SELECT * FROM resources where resource_type = ?", ("Procedure",)
        )

        for _ in cursor.fetchall():
            key, resource_type, procedure = _
            procedure = json.loads(procedure)
            # simplify the identifier
            procedure["identifier"] = procedure["identifier"][0]["value"]
            # simplify the code
            procedure["code"] = procedure["code"]["coding"][0]["display"]
            # simplify the reason
            procedure["reason"] = procedure["reason"][0]["reference"]["reference"]
            # simplify the occurrenceAge
            procedure["occurrenceAge"] = procedure["occurrenceAge"]["value"]
            # simplify the subject
            subject = procedure["subject"]["reference"]
            procedure["subject"] = subject

            if subject.startswith("Patient/"):
                _, patient_id = subject.split("/")
                resources = [_ for _ in loaded_db.patient_everything(patient_id)]
                resources.append(loaded_db.patient(patient_id))
                for resource in resources:

                    if resource["resourceType"] == "Patient":
                        procedure["patient"] = resource["identifier"][0]["value"]
                        continue

                    if (
                        resource["resourceType"] == "Condition"
                        and f"Condition/{resource['id']}" == procedure["reason"]
                    ):
                        procedure["reason"] = resource["code"]["text"]
                        continue

                    if resource["resourceType"] == "Observation":
                        # must be focus
                        if f"Procedure/{procedure['id']}" not in [
                            _["reference"] for _ in resource["focus"]
                        ]:
                            continue

                        # TODO - pick first coding, h2 allow user to specify preferred coding
                        code = resource["code"]["coding"][0]["code"]

                        if "valueQuantity" in resource:
                            value = resource["valueQuantity"]["value"]
                        elif "valueCodeableConcept" in resource:
                            value = resource["valueCodeableConcept"]["text"]
                        elif "valueInteger" in resource:
                            value = resource["valueInteger"]
                        elif "valueString" in resource:
                            value = resource["valueString"]
                        else:
                            value = None

                        assert value is not None, f"no value for {resource['id']}"
                        procedure[code] = value

                        continue

                    # skip these
                    if resource["resourceType"] in [
                        "Specimen",
                        "Procedure",
                        "ResearchSubject",
                    ]:
                        continue

                    # default, add entire resource as an item of the list
                    resource_type = inflection.underscore(resource["resourceType"])
                    if resource_type not in procedure:
                        procedure[resource_type] = []
                    procedure[resource_type].append(resource)

            yield procedure

    def handle_units(self, value_normalized: str):
        """This function is designed to attempt to remove units string suffixes
        and attempt to store values as float. The issue arises when elastic sees
        string and float data in the same column and gives errors because it is expecting
        only one data type per column"""

        if value_normalized is not None:
            value_normalized_split = value_normalized.split(" ")
            if isinstance(value_normalized_split, list):
                value_numeric = value_normalized_split[0]
                if is_number(value_numeric):
                    value_normalized = float(value_numeric)
            return value_normalized
        return None

    def select_category(self, resource):
        if get_nested_value(resource, ["category", 0, "coding", 0]):
            selected_coding = None

            # Loop through each coding entry
            for coding in resource["category"][0]["coding"]:
                # Check if coding system is SNOMED CT
                if coding.get("system") == "http://snomed.info/sct":
                    selected_coding = coding["display"]
                    break  # Found SNOMED code, exit loop

                # If SNOMED code not found, select the first coding
                if selected_coding is None:
                    selected_coding = coding["display"]

            # Now selected_coding contains the desired coding entry
            # Proceed with further actions or return selected_coding
            return selected_coding
        else:
            # Handle case where no coding entries exist
            return None

    def select_coding(self, resource):
        """
        Selects and returns a coding entry from a FHIR Condition resource based on the following priority:
        1. Selects a coding entry with system 'http://snomed.info/sct' if present.
        2. If no SNOMED coding is found, selects the first coding entry.

        Args:
            condition (dict): The FHIR Condition resource containing 'code' and 'code.coding' fields.

        Returns:
            dict or None: The selected coding entry dictionary if found, or None if no coding entries exist.
        """
        if get_nested_value(resource, ["code", "coding", 0]):
            selected_coding = None

            # Loop through each coding entry
            for coding in resource["code"]["coding"]:
                # Check if coding system is SNOMED CT
                if coding.get("system") == "http://snomed.info/sct":
                    selected_coding = coding["display"]
                    break  # Found SNOMED code, exit loop

                # If SNOMED code not found, select the first coding
                if selected_coding is None:
                    selected_coding = coding["display"]

            # Now selected_coding contains the desired coding entry
            # Proceed with further actions or return selected_coding
            return selected_coding
        else:
            # Handle case where no coding entries exist
            return None

    def flattened_research_subjects(self) -> Generator[dict, None, None]:

        # get all observations with a Observation.subject=Patient, mapped from patient ID to observation
        resource_type = "ResearchSubject"
        conditions_by_patient_id = get_conditions_by_subject(self, "Patient")

        # get all ResearchSubjects
        cursor = self.connect()
        cursor.execute(
            "SELECT * FROM resources where resource_type = ?", (resource_type,)
        )

        # get research subject and associated .subject patient
        for _, _, raw_research_subject in cursor.fetchall():
            research_subject = json.loads(raw_research_subject)
            flat_research_subject = SimplifiedResource.build(
                resource=research_subject
            ).simplified

            # return with .subject (ie Patient) fields
            patient = get_subject(self, research_subject)
            flat_research_subject.update(patient)

            # get condition code, eg enrollment diagnosis
            if patient["patient_id"] in conditions_by_patient_id:
                conditions = conditions_by_patient_id[patient["patient_id"]]

                # TODO: assumes there are no duplicate column names in each condition
                for condition in conditions:
                    for k, v in traverse(condition).items():
                        if k not in set(["condition_id", "condition_identifier"]):
                            flat_research_subject[k] = v

            yield flat_research_subject

    def flattened_document_references(self) -> Generator[dict, None, None]:
        """generator that yields document references populated
        with DocumentReference.subject fields and Observation codes through Observation.focus
        """

        cursor = self.connect()
        resource_type = "DocumentReference"

        # get a dict mapping focus ID to its associated observations
        observation_by_focus_id = get_observations_by_focus(self, resource_type)

        # flatten each document reference
        cursor.execute(
            "SELECT * FROM resources where resource_type = ?", (resource_type,)
        )
        for _, _, resource in cursor.fetchall():
            document_reference = json.loads(resource)
            yield self.flattened_document_reference(
                document_reference, observation_by_focus_id
            )

    def flattened_document_reference(
        self, doc_ref: dict, observation_by_focus_id: dict
    ) -> dict:
        # simplify document reference
        flat_doc_ref = SimplifiedResource.build(resource=doc_ref).simplified

        # extract the corresponding .subject and append its fields
        flat_doc_ref.update(get_subject(self, doc_ref))

        # populate observation data associated with the document reference document
        if doc_ref["id"] in observation_by_focus_id:
            associated_observations = observation_by_focus_id[doc_ref["id"]]

            # TODO: assumes there are no duplicate column names in each observation
            for observation in associated_observations:
                flat_observation = SimplifiedResource.build(resource=observation).values
                flat_doc_ref.update(flat_observation)

        # TODO: test this based on fhir-gdc
        if "basedOn" in doc_ref:
            for i, dict_ in enumerate(doc_ref["basedOn"]):
                doc_ref[f"basedOn_{i}"] = dict_["reference"]
            del doc_ref["basedOn"]

        return flat_doc_ref

    @lru_cache(maxsize=None)
    def flattened_specimens(self) -> Generator[dict, None, None]:
        """generator that yields specimens populated with Specimen.subject fields
        and Observation codes through Observation.focus"""

        resource_type = "Specimen"
        cursor = self.connect()

        # get a dict mapping focus ID to its associated observations
        observations_by_focus_id = get_observations_by_focus(self, resource_type)

        # flatten each document reference
        cursor.execute(
            "SELECT * FROM resources where resource_type = ?", (resource_type,)
        )
        for _, _, resource in cursor.fetchall():
            specimen = json.loads(resource)
            yield self.flattened_specimen(specimen, observations_by_focus_id)

    def flattened_specimen(self, specimen: dict, observation_by_id: dict) -> dict:
        """Return the specimen with everything resolved."""

        # create simple specimen dict
        flat_specimen = SimplifiedResource.build(resource=specimen).simplified

        # extract its .subject and append its fields (including id)
        flat_specimen.update(get_subject(self, specimen))

        # populate observation codes for each associated observation
        if specimen["id"] in observation_by_id:
            observations = observation_by_id[specimen["id"]]

            # TODO: assumes there are no duplicate column names in each observation
            for observation in observations:
                flat_observation = SimplifiedResource.build(resource=observation).values
                flat_specimen.update(flat_observation)

        return flat_specimen


def create_dataframe(
    directory_path: str, work_path: str, data_type: str
) -> pd.DataFrame:
    """Create a dataframe from the FHIR data in the directory."""
    assert pathlib.Path(work_path).exists(), f"Directory {work_path} does not exist."
    work_path = pathlib.Path(work_path)
    db_path = work_path / "local_fhir.db"
    db_path.unlink(missing_ok=True)

    db = LocalFHIRDatabase(db_name=db_path)
    db.load_ndjson_from_dir(path=directory_path)

    if data_type == "DocumentReference":
        df = pd.DataFrame(db.flattened_document_references())
    elif data_type == "ResearchSubject":
        df = pd.DataFrame(db.flattened_research_subjects())
    elif data_type == "Specimen":
        df = pd.DataFrame(db.flattened_specimens())
    else:
        raise ValueError(
            f"{data_type} not supported yet. Supported data types are DocumentReference, ResearchSubject, and Specimen"
        )
    assert (
        not df.empty
    ), "Dataframe is empty, are there any DocumentReference resources?"

    front_column_names = ["resourceType", "identifier"]
    if "patient" in df.columns:
        front_column_names = front_column_names + ["patient"]

    remaining_columns = [col for col in df.columns if col not in front_column_names]
    rear_column_names = [
        "id"
    ]  # removed status for the purpose of not needing it for the demo
    if "subject" in df.columns:
        rear_column_names = rear_column_names + ["subject"]
    for c in df.columns:
        if c.endswith("_identifier"):
            rear_column_names.append(c)
    remaining_columns = [
        col for col in remaining_columns if col not in rear_column_names
    ]

    reordered_columns = front_column_names + remaining_columns + rear_column_names
    df = df[reordered_columns]
    df = df.replace({np.nan: ""})
    return df


def is_number(s):
    """Returns True if string is a number."""
    try:
        int(s)
        return True
    except ValueError:
        return False


def get_subject(db: LocalFHIRDatabase, resource: dict) -> dict:
    """get the resource's subject field if it exists"""

    # ensure resource has subject field
    subject_key = get_nested_value(resource, ["subject", "reference"])
    if subject_key is None:
        return {}

    # traverse the resource of the subject and return its values
    cursor = db.connect()
    cursor.execute("SELECT * FROM resources WHERE key = ?", (subject_key,))
    row = cursor.fetchone()
    assert row, f"{subject_key} not found in database"
    _, _, raw_subject = row
    subject = json.loads(raw_subject)
    return traverse(subject)


def get_resources_by_reference(
    db: LocalFHIRDatabase, resource_type: str, reference_field: str, reference_type: str
) -> dict[str, list]:
    """given a set of rescode ources of type resource_type, map each unique reference in reference field of type reference_type to its associated resources
    ex: use all Observations with a Specimen focus, map Specimen IDs to its list of associated Observations and return the map
    """

    # ensure reference field is allowed
    allowed_fields = ["focus", "subject"]
    assert (
        reference_field in allowed_fields
    ), f"Field not implemented, choose between {allowed_fields}"

    cursor = db.connect()
    cursor.execute(
        """
        SELECT *
        FROM resources
        WHERE resource_type = ?
    """,
        (resource_type,),
    )

    resource_by_reference_id = defaultdict(list)

    for _, _, raw_resource in cursor.fetchall():
        resource = json.loads(raw_resource)

        # determine which how to process the field
        if reference_field == "focus":
            # error if multiple focuses
            #if reference_field in resource:
                #assert (
                #    len(resource["focus"]) <= 1
                #), "unable to support more than 1 focus for a single observation"
            nested_keys = ["focus", 0]
        elif reference_field == "subject":
            nested_keys = ["subject"]

        # add observation to dict if a reference resource exists
        reference_key = get_nested_value(resource, [*nested_keys, "reference"])
        if reference_key is not None and reference_type in reference_key:
            reference_id = reference_key.split("/")[-1]
            resource_by_reference_id[reference_id].append(resource)

    return resource_by_reference_id


def get_observations_by_focus(
    db: LocalFHIRDatabase, focus_type: str
) -> dict[str, list]:
    """get all Observations that have a focus of resource type focus_type"""
    return get_resources_by_reference(db, "Observation", "focus", focus_type)


def get_conditions_by_subject(
    db: LocalFHIRDatabase, subject_type: str
) -> dict[str, list]:
    """get all Conditions that have a subject of resource type subject_type"""
    return get_resources_by_reference(db, "Condition", "subject", subject_type)



import logging


def ls(config, object_id: str = None, metadata: dict = {}, auth=None):
    """List files."""
    from gen3_tracker import Config
    config: Config = config
    from gen3.index import Gen3Index
    index_client = Gen3Index(auth_provider=auth)

    if object_id:
        if ',' in object_id:
            object_ids = object_id.split(',')
        else:
            object_ids = [object_id]
        records = index_client.client.bulk_request(dids=object_ids)
        return {'records': [_.to_json() for _ in records]}

    params = {}
    project_id = metadata.get('project_id', None)
    if 'project_id' in metadata:
        program, project = project_id.split('-')
        params = {'authz': f"/programs/{program}/projects/{project}"}
        metadata.pop('project_id')
    params['metadata'] = metadata

    records = index_client.client.list_with_params(params=params)

    def _ensure_project_id(record):
        if 'project_id' not in record['metadata'] and project_id:
            record['metadata']['project_id'] = project_id
        return record

    records = [_ensure_project_id(_.to_json()) for _ in records]

    return {
        'records': records,
        'msg': f"Project {project_id} has {len(records)} files."
    }


def find_latest_snapshot(auth, config):
    """
    Find the latest snapshot for a project.
    Looks for a hierarchy of files in the indexd database and returns the latest one found. The hierarchy is:
        * the latest git snapshot
        * the latest SNAPSHOT.zip created by the fhir-import-export job on output
        * the latest meta.zip created by the fhir-import-export client on input
    """
    results = ls(config=config, metadata={'project_id': config.gen3.project_id}, auth=auth)
    records = 'records' in results and results['records'] or []
    file_names = [_['file_name'] for _ in records]
    git_records = [r for r in records if 'git' in r['file_name']]
    git_records = sorted(git_records, key=lambda d: d['file_name'])
    download_meta = None
    if len(git_records) > 0:
        # most recent metadata, file_name has a timestamp
        download_meta = git_records[-1]
    else:
        logger = logging.getLogger(__name__)
        logger.info(f"No git snapshot found for {config.gen3.project_id}")
        snapshot_records = [r for r in records if 'SNAPSHOT.zip' in r['file_name']]
        snapshot_records = sorted(snapshot_records, key=lambda d: d['file_name'])
        if len(snapshot_records) > 0:
            download_meta = snapshot_records[-1]
        else:
            logger.info(f"No SNAPSHOT found for {config.gen3.project_id}")
            meta_records = [r for r in records if 'meta.zip' in r['file_name']]
            meta_records = sorted(meta_records, key=lambda d: d['file_name'])
            if len(meta_records) > 0:
                download_meta = meta_records[-1]

    assert download_meta, f"No git, snapshot or meta files found for {config.gen3.project_id}, file_names: {file_names}"
    return download_meta



import click

from gen3_tracker.common import validate_email
from gen3_tracker.config import Config, ensure_auth
from gen3_tracker.common import CLIOutput
from gen3_tracker import NaturalOrderGroup


@click.group(name='access', cls=NaturalOrderGroup)
@click.pass_obj
def access_group(config: Config):
    """Manage access requests."""
    pass


@access_group.command(name="add")
@click.argument('user_name')
@click.option('--resource_path', default=None, required=False, show_default=True,
              help="Gen3 authz /programs/<program>")
@click.option('--roles', show_default=True, default=None, help='Add comma-delimited role permissions to the access request, ex: --roles "storage_writer,file_uploader"')
@click.option('--steward', show_default=True, is_flag=True, default=False, help='Add steward role to the program')
@click.pass_obj
def access_touch(config: Config, resource_path: str, user_name: str, roles: str, steward: bool):
    """Create a request a specific role.

    \b
    USER_NAME (str): user's email

    """
    from gen3_tracker.collaborator.access import create_request

    with CLIOutput(config=config) as output:
        try:

            msgs = validate_email(user_name)
            assert msgs == [], f"Invalid email address: {user_name} {msgs}"

            assert user_name, "user_name required"

            request = {"username": user_name, "resource_path": resource_path}
            if steward:
                roles = 'requestor_reader_role,requestor_updater_role'

            assert roles, "roles required"
            roles = roles.split(',')
            for role in roles:
                request.update({"role_ids": [role]})
                print(request)
                output.update(create_request(config=config, request=request))
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1


@access_group.command(name="sign")
@click.option('--username', required=False, help='Sign all requests for user within a project')
@click.option('--request_id', required=False, help='Sign only this request')
@click.pass_obj
def sign(config: Config, username: str, request_id: str):
    """Sign all policies for a project.
    \b
    """
    with CLIOutput(config=config) as output:
        from gen3_tracker.collaborator.access.requestor import ls, update, LogAccess

        auth = ensure_auth(config=config)
        access = ls(config, mine=False, username=username, active=True, auth=auth)
        unsigned_requests = [_ for _ in access.requests if _['status'] != 'SIGNED']

        if len(unsigned_requests) == 0:
            output.update(LogAccess(**{
                'msg': "No unsigned requests found"
            }))
        else:
            msg = f"Signing {len(unsigned_requests)} requests."

            signed_requests = []
            problem_requests = []
            click.secho("signing requests...", fg='green')
            for request in unsigned_requests:
                if request_id and request['request_id'] != request_id:
                    continue
                try:
                    signed_requests.append(
                        update(config, request_id=request['request_id'], status='SIGNED', auth=auth).request
                    )
                except Exception as e:
                    print(f"Error signing request {request}: {e}")
                    problem_requests.append(request)

            msg = f"Signed {len(signed_requests)} requests. There were {len(problem_requests)} problems. System administrators will create new projects."

            output.update(LogAccess(**{
                'msg': msg,
                'requests': signed_requests,
            }))


@access_group.command(name="ls")
@click.option('--username', required=False, default=None, help='Sign all requests for user within a project')
@click.option('--mine', is_flag=True, show_default=True, default=False, help="List current user's requests. Otherwise, list all the requests the current user has access to see.")
@click.option('--all', 'active', is_flag=True, show_default=True, default=True, help='Only unsigned requests')
@click.pass_obj
def access_ls(config: Config, mine: bool, active: bool, username: str):
    """List current user's requests."""
    from gen3_tracker.collaborator.access.requestor import ls

    with CLIOutput(config=config) as output:
        try:
            msg = 'OK'
            access = ls(config, mine, active, username)
            if not access.requests or len(access.requests) == 0:
                msg = 'No unsigned requests'
            output.update({'requests': access.requests, 'msg': msg})
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1


@access_group.command(name="cat")
@click.argument('request_id')
@click.pass_obj
def access_cat(config: Config, request_id: str):
    """Show details of a specific request.

    \b
    request_id (str): uuid of an existing request
    """
    from gen3_tracker.collaborator.access.requestor import cat

    with CLIOutput(config=config) as output:
        output.update(cat(config, request_id))



import json
import pathlib
import subprocess
import sys
import typing
import uuid
from collections import OrderedDict
from typing import Union, Optional

import click
import pydantic
from click import Context, Command
from pydantic import BaseModel, field_validator


ACED_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, b'aced-idp.org')
ENV_VARIABLE_PREFIX = 'G3T_'

FILE_TRANSFER_METHODS = {
    'gen3': 'gen3-client to/from local',
    'no-bucket': 'indexd only, symlink to/from local',
    's3': '(admin) s3 to/from local',
    's3-map': '(admin) s3 index only external s3',
}


def monkey_patch_url_validate():
    # monkey patch to allow file: urls
    import fhir.resources.fhirtypes
    from pydantic import FileUrl

    original_url_validate = fhir.resources.fhirtypes.Url.validate

    @classmethod
    def better_url_validate(cls, value: str, field: "ModelField", config: "BaseConfig") -> Union["AnyUrl", str]:    # noqa
        """Allow file: urls. see https://github.com/pydantic/pydantic/issues/1983
        bugfix: addresses issue introduced with `fhir.resources`==7.0.1
        """
        if value.startswith("file:"):
            _ = FileUrl(value)
            return value
            # return FileUrl.validate(value, field, config)
        value = original_url_validate(value, field, config)
        return value

    fhir.resources.fhirtypes.Url.validate = better_url_validate


class LogConfig(BaseModel):
    format: str
    """https://docs.python.org/3/library/logging.html#logging.Formatter"""
    level: str
    """https://docs.python.org/3/library/logging.html#logging-levels"""


class OutputConfig(BaseModel):
    format: str = "text"
    """write to stdout with this format"""


class _DataclassConfig:
    """Pydantic dataclass configuration

    See https://docs.pydantic.dev/latest/usage/model_config/#options"""
    arbitrary_types_allowed = True


class Gen3Config(BaseModel):

    profile: str = None
    """The name of the gen3-client profile in use. See https://bit.ly/3NbKGi4"""

    version: Optional[str] = None
    """The version of gen3-client in use."""

    project_id: str = None
    """The program-project."""

    @field_validator('project_id')
    def check_project_id(cls, v):
        if v.count('-') != 1:
            raise ValueError('project_id must contain exactly one "-"')
        return v

    @property
    def program(self) -> str:
        if not self.project_id:
            return None
        return self.project_id.split('-')[0]

    @property
    def project(self) -> str:
        if not self.project_id:
            return None
        return self.project_id.split('-')[1]

    @property
    def authz(self) -> str:
        if not self.project_id:
            return None
        return f'/programs/{self.program}/projects/{self.project}'


class Config(BaseModel):
    log: LogConfig = LogConfig(
        format='[%(asctime)s] — [%(levelname)s] — %(name)s — %(message)s',
        level='INFO'
    )
    """logging setup"""
    output: OutputConfig = OutputConfig(format='yaml')
    """output setup"""
    gen3: Gen3Config = Gen3Config()
    """gen3 setup"""
    work_dir: pathlib.Path = None
    """temporary files"""
    no_config_found: bool = False
    """Is this default config used because none found in cwd or parents?"""
    debug: bool = False
    """Enable debug mode, lots of logs."""
    dry_run: bool = False
    """Print the commands that would be executed, but do not execute them."""

    def model_dump(self):
        """Dump the config model.

         temporary until we switch to pydantic2
        """
        _ = json.loads(self.model_dump_json())
        del _['no_config_found']
        return _

    def commit_dir(self):
        """Return the path to the commits' directory."""
        return self.state_dir / self.gen3.project_id / 'commits'


# cli helpers -------------------------------------------------------------------
class NaturalOrderGroup(click.Group):
    """Allow listing Commands in order of appearance, with common parameters."""
    # see https://github.com/pallets/click/issues/513

    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name,
                             commands=commands,
                             **attrs)

    def list_commands(self, ctx):
        # print('list_commands', self.commands.keys())
        return self.commands.keys()

    def resolve_command(
        self, ctx: Context, args: typing.List[str]
    ) -> typing.Tuple[typing.Optional[str], typing.Optional[Command], typing.List[str]]:
        # print('resolve_command', args)
        try:
            return super().resolve_command(ctx, args)
        except Exception as e:
            if 'No such command' in str(e):
                # delegate to git
                try:
                    from gen3_tracker.git import run_command
                    result = run_command(f'git {" ".join(args)}', dry_run=False, no_capture=True)
                    sys.exit(result.return_code)
                    # os._exit(result.return_code)  # noqa
                except subprocess.CalledProcessError as e2:
                    # os._exit(e2.returncode)  # noqa
                    sys.exit(e2.returncode)

                # # suggest git prompt
                # from g3t.common import ERROR_COLOR
                # click.secho(f'Command not found: `g3t {" ".join(args)}`, did you mean: `git {" ".join(args)}` ?',
                #             fg=ERROR_COLOR, file=sys.stderr)
                # os._exit(1)  # noqa

            raise e


# main
monkey_patch_url_validate()

# default initializers for path
pydantic.v1.json.ENCODERS_BY_TYPE[pathlib.PosixPath] = str
pydantic.v1.json.ENCODERS_BY_TYPE[pathlib.WindowsPath] = str
pydantic.v1.json.ENCODERS_BY_TYPE[pathlib.Path] = str



import json
import logging
import mimetypes
import multiprocessing
import os
import pathlib
import subprocess
import time
import typing
import zipfile
from abc import abstractmethod
from datetime import datetime
from typing import NamedTuple

import inflection
import pydantic
import pytz
import yaml
from fhir.resources.attachment import Attachment
from gen3.auth import Gen3Auth
from pydantic import BaseModel, ConfigDict, field_validator

from gen3_tracker.common import ACCEPTABLE_HASHES, parse_iso_tz_date, create_object_id

# constants ---------------------------------------------------------------------
INIT_MESSAGE = 'Initializing a new repository...'
ADD_MESSAGE = 'Adding files to the repository...'
COMMIT_MESSAGE = 'Committing changes to the repository...'
PUSH_MESSAGE = 'Pushing changes to the remote repository...'
RESET_MESSAGE = 'Resetting changes in the repository...'
UNDO_INIT_MESSAGE = 'Removing change control from repository...'
MISSING_GIT_MESSAGE = '.git is not initialized in the current directory.'
MISSING_G3T_MESSAGE = '.g3t is not initialized in the current directory.'
STATUS_MESSAGE = 'Showing changed stages...'
DIFF_MESSAGE = 'Showing details of changed files...'

LOGGED_ALREADY = set()

mimetypes.add_type('text/fastq', '.fastq')
mimetypes.add_type('text/fastq', '.fq')

# process helpers ---------------------------------------------------------------


class CommandResult(NamedTuple):
    """
    A NamedTuple that represents the result of a command execution.

    Attributes:
        stdout (str): The standard output of the command.
        stderr (str): The standard error of the command.
        return_code (int): The return code of the command.
    """
    stdout: str
    stderr: str
    return_code: int


# https://dvc.org/doc/user-guide/project-structure/dvc-files#dvc-files
class DVCItem(BaseModel):
    hash: str

    md5: typing.Optional[str] = None
    sha256: typing.Optional[str] = None
    sha1: typing.Optional[str] = None
    sha512: typing.Optional[str] = None
    crc: typing.Optional[str] = None
    etag: typing.Optional[str] = None

    mime: typing.Optional[str] = 'application/octet-stream'
    modified: str
    path: str
    size: int
    is_symlink: typing.Optional[bool] = False
    realpath: typing.Optional[str] = None
    """For symlinked files, the real path to the file."""
    source_url: typing.Optional[str] = None
    """bucket imports, the url to the file."""

    object_id: typing.Optional[str] = None

    @pydantic.model_validator(mode="after")
    def check_hash_value(self):
        """Check that the hash value is valid."""
        hash_type = self.hash
        assert hash_type in ACCEPTABLE_HASHES, f'Invalid hash type {hash_type}'
        v = getattr(self, hash_type)
        assert ACCEPTABLE_HASHES[hash_type](v), f'Invalid {hash_type} {v}'
        self.modified = parse_iso_tz_date(self.modified)
        return self

    @property
    def hash_value(self):
        """Get the hash value."""
        return getattr(self, self.hash)

    def set_object_id(self, project_id: str) -> str:
        """ create a unique did for this object within a project"""
        assert self.path, 'path is required'

        self.object_id = create_object_id(self.path, project_id)
        return self.object_id


class DVCMeta(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )

    patient: typing.Optional[str] = None
    specimen: typing.Optional[str] = None
    observation: typing.Optional[str] = None
    task: typing.Optional[str] = None
    no_bucket: typing.Optional[bool] = False


class DVC(BaseModel):
    model_config = ConfigDict(
        extra='allow',
    )
    meta: typing.Optional[DVCMeta] = None
    outs: list[DVCItem]
    project_id: typing.Optional[str] = None

    @field_validator('outs')
    def check_outs(cls, v):
        if len(v) < 1:
            raise ValueError('outs must contain at least one item')
        return v

    @pydantic.model_validator(mode="after")
    def check_object_id(self):
        """Set the object_id if empty."""
        if not self.object_id and self.project_id:
            self.outs[0].set_object_id(self.project_id)
        return self

    @property
    def out(self) -> DVCItem:
        """Convenience method to get the first out."""
        return self.outs[0]

    @property
    def object_id(self) -> str:
        """Convenience method to get the first object_id."""
        if not self.outs[0].object_id and self.project_id:
            self.outs[0].set_object_id(self.project_id)
        return self.outs[0].object_id

    @classmethod
    def from_document_reference(cls, config, document_reference, references) -> 'DVC':
        """Factory: create DVC from DocumentReference.

        Args:
            config (object): our config object
            document_reference (DocumentReference): loaded DocumentReferences
            references (dict): cross ref of Reference to Identifiers

        """
        attachment: Attachment = document_reference.content[0].attachment
        assert attachment.extension, document_reference
        source_path = next(iter([_.valueUrl for _ in attachment.extension if _.url.endswith('source_path')]), None)
        if not source_path:
            source_path = attachment.url

        for k in ACCEPTABLE_HASHES.keys():
            if [_.valueString for _ in attachment.extension if _.url.endswith(k)]:
                hash_value = [_.valueString for _ in attachment.extension if _.url.endswith(k)][0]
                hash_name = k
                break

        assert source_path, ("missing source_path", document_reference)
        assert hash_value, ("missing hash_value", document_reference)

        resource_type, resource_id = document_reference.subject.reference.split('/')
        meta = DVCMeta()

        if not resource_type == 'ResearchStudy':
            resource_type = inflection.underscore(resource_type)
            identifier = references[document_reference.subject.reference]
            meta = DVCMeta(**{resource_type: identifier})

        modified = None
        if attachment.creation:
            modified = attachment.creation.isoformat()
        elif document_reference.date:
            modified = document_reference.date.isoformat()

        assert modified, ("missing attachment.creation, document_reference.date", document_reference)

        dvc_object = DVC(
            project_id=config.gen3.project_id,
            meta=meta,
            outs=[
                DVCItem(
                    modified=modified,
                    path=attachment.url.replace('file:///', ''),
                    size=attachment.size,
                    realpath=source_path.replace('file:///', ''),
                    mime=attachment.contentType,
                    object_id=document_reference.id,
                    **{hash_name: hash_value, 'hash': hash_name}
                )
            ],
        )

        if document_reference.subject:

            if document_reference.subject.reference in references:
                key = document_reference.subject.reference.split('/')[0]

                if key not in ['ResearchStudy']:
                    value = references[document_reference.subject.reference]
                    meta = DVCMeta.model_validate({key: value})
                    assert isinstance(meta, DVCMeta), f"Did not get expected meta {meta}"
                    dvc_object.meta = meta
            else:
                msg = f"Did not find {document_reference.subject.reference} in references"
                if msg not in LOGGED_ALREADY:
                    logging.getLogger(__package__).warning(msg)
                    LOGGED_ALREADY.add(msg)
        assert attachment.url.replace('file:///', '') == dvc_object.out.path, "attachment and dvc path doesn't match"
        assert document_reference.id == dvc_object.object_id, f"Did not get expected ID {config.gen3.project_id} {attachment.url}/{document_reference.id} {dvc_object.out.path}/{dvc_object.object_id}"
        return dvc_object


def run_command(command: str, dry_run: bool = False, raise_on_err: bool = True, env: dict = None, no_capture: bool = False) -> CommandResult:
    """Run a shell command and return its output. Raise an exception if the command fails."""
    _logger = logging.getLogger(__package__)
    if dry_run:
        _logger.info(f'dry_run: {command}')
        return CommandResult('', '', 0)
    else:
        if no_capture:
            process = subprocess.run(command, shell=True, env=env, check=True)
            stdout = b''
            stderr = b''
        else:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, env=env)
            stdout, stderr = process.communicate()

        if raise_on_err and process.returncode != 0:
            raise Exception(f'Command `{command}` failed with error code {process.returncode}, stderr: {stderr.decode()}')
        rc = CommandResult(stdout.decode(), stderr.decode(), process.returncode)
        _logger.debug(f'Command `{command}` returned {rc}')
        return rc


# git helpers -------------------------------------------------------------------
def git_remote_exists(dry_run: bool = False) -> bool:
    """Check if git repo has a remote."""

    stdout, stderr, return_code = run_command('git remote -v', dry_run=dry_run, raise_on_err=False)
    if '' != stdout:
        print('git_remote_exists', f">{stdout}<")
        return True
    return False


def git_repository_exists(dry_run: bool = False) -> bool:
    """Check if git is already initialized"""

    stdout, stderr, return_code = run_command('git status', dry_run=dry_run, raise_on_err=False)
    if return_code != 0:
        return False
    return True


def git_status() -> dict:
    """Get the status of the git repository"""
    result = run_command('git status -s --branch')
    assert result.return_code == 0, result.stderr
    path_statuses = []
    for line in result.stdout.split('\n'):
        if not line:
            continue
        if line.startswith('##'):
            branch = line.split(' ')[1]
            continue
        if not line.startswith('??'):
            _status, path = line.split()
            path_statuses.append({'status': _status, 'path': path})
            continue
    return branch, path_statuses


def git_ls(dry_run: bool = False) -> list[dict]:
    """List the files in the git repository"""
    results = run_command('git ls-files', dry_run=dry_run)
    return json.loads(results.stdout)


def git_files(dry_run=False) -> list[str]:
    """List the files in the git repository"""
    # """
    # %h = abbreviated commit hash
    # %x09 = tab (character for code 9)
    # %an = author name
    # %ad = author date (format respects --date= option)
    # %s = subject
    # """
    # get all files in the repository
    result = run_command('git log --pretty=format:"%h%x09%an%x09%ad%x09%s"  --date=iso-strict ', dry_run=dry_run, raise_on_err=False)
    if result.return_code == 0:
        git_logs = [_.split('\t') for _ in result.stdout.split('\n')]
        git_logs = [{'hash': _[0], 'author': _[1], 'timestamp': _[2], 'subject': _[3]} for _ in git_logs]
        for _ in git_logs:
            # diff = run_command(f'git diff --name-only {_["hash"]}~1 {_["hash"]}', ctx.obj['DRY_RUN'])
            files = run_command(f'git ls-tree --name-only -r {_["hash"]}', dry_run=dry_run)
            _['files'] = [_ for _ in files.stdout.split('\n') if _ != '']
        # de-duplicate the files
        to_upload = set()
        for _ in git_logs:
            to_upload.update([_ for _ in _['files'] if _.startswith('MANIFEST')])
        return list(to_upload)
    return []


# file helpers ------------------------------------------------------------------
def get_mime_type(file_name):
    """Get mime type from file name."""
    return mimetypes.guess_type(file_name, strict=False)[0] or 'application/octet-stream'


def calculate_hash(hash_type, file_name):
    """Calculate the MD5 hash of a file."""
    assert hash_type == 'md5', f'Hash type {hash_type} is not supported (yet'
    import hashlib
    hash_md5 = hashlib.md5()
    with open(file_name, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def git_archive(zip_name):
    """Archive the current branch and it's content to a zip file."""
    result = run_command("git rev-parse --abbrev-ref HEAD", no_capture=False)
    assert result.return_code == 0, f"Could not get current branch {result.stderr}"
    branch = result.stdout.strip()
    assert branch, "Could not get current branch"
    # add all the content
    run_command(f"git archive -o {zip_name} {branch}", no_capture=False)
    # add the .git folder
    with zipfile.ZipFile(zip_name, 'a') as zipf:  # 'a' for append mode
        for root, dirs, files in os.walk('.git'):
            for file in files:
                zipf.write(os.path.join(root, file),
                           os.path.relpath(os.path.join(root, file),
                           os.path.join('.git', '..')))


def modified_date(file_path):
    return datetime.fromtimestamp(os.path.getmtime(file_path), pytz.UTC).isoformat()


# manifest helpers --------------------------------------------------------------

class ManifestChange(NamedTuple):
    """A NamedTuple that represents a change in the manifest file."""
    data_path: pathlib.Path
    dvc_path: pathlib.Path


def data_file_changes(manifest_path, update: bool = False) -> list[ManifestChange]:
    """Check for changes in the dvc files timestamps, return the data path and the dvc file"""
    changes = []
    for _ in manifest_path.rglob('*.dvc'):
        dvc = to_dvc(_)

        if not dvc.out.realpath or dvc.out.source_url:
            continue

        data_path = pathlib.Path(dvc.out.realpath)

        if not data_path.is_symlink() and not data_path.exists():
            continue

        if data_path.stat().st_mtime > _.stat().st_mtime:
            if update:
                with open(_, 'r') as f:
                    yaml_data = yaml.safe_load(f)
                    out = yaml_data['outs'][0]
                    out['size'] = data_path.stat().st_size
                    out['modified'] = datetime.fromtimestamp(data_path.stat().st_mtime, pytz.UTC).isoformat()
                    out[out['hash']] = calculate_hash(out['hash'], data_path)
                    yaml_data['outs'] = [out]
                with open(_, 'w') as f:
                    yaml.dump(yaml_data, f, default_flow_style=False)
            #
            # return the data path and the dvc file
            #
            changes.append(ManifestChange(data_path, _))
    return changes


# meta data helpers ------------------------------------------------------------

def update_meta(file_path, new_meta):
    # Check if file exists
    # lazy loading improves startup
    from deepdiff import DeepDiff

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"{file_path} does not exist.")

    # Get file attributes
    file_stat = os.stat(file_path)
    file_attributes = {
        'size': file_stat.st_size,
        'modified': file_stat.st_mtime,
    }

    # Load existing data from the YAML file
    with open(file_path, 'r') as f:
        yaml_data = yaml.safe_load(f)

    # Check if file attributes in the YAML data match the actual file attributes
    if yaml_data['outs'][0]['size'] == file_attributes['size'] and \
       yaml_data['outs'][0]['modified'] == file_attributes['modified']:
        # If file attributes haven't changed, check if the metadata has changed
        if 'meta' in yaml_data and DeepDiff(yaml_data['meta'], new_meta, ignore_order=True):
            # If metadata has changed, update it
            yaml_data['meta'] = new_meta

            # Write the updated data back to the YAML file
            with open(file_path, 'w') as f:
                yaml.dump(yaml_data, f, default_flow_style=False)
            return True

    return False


# gen3 helpers ------------------------------------------------------------------

def dvc_data(committed_files) -> typing.Generator[DVC, None, None]:
    """Get the dvc data from the committed files."""
    for committed_file in committed_files:
        if str(committed_file).endswith('.dvc'):
            yield to_dvc(committed_file)


def to_dvc(path) -> DVC:
    """Get the dvc data from a file."""
    with open(path, 'r') as f:
        _ = yaml.safe_load(f)
        return DVC.model_validate(_)


class LoggingWriter:
    def __init__(self, log_file=None):
        assert log_file is not None, 'log_file is required'
        pathlib.Path(log_file).parent.mkdir(parents=True, exist_ok=True)
        pathlib.Path(log_file).unlink(missing_ok=True)
        self.log_file = log_file
        self.logger = None

    def __enter__(self):
        self.logger = logging.getLogger('LoggingWriter')
        handler = logging.FileHandler(self.log_file)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        # Prevent logging to console
        self.logger.propagate = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)

    @abstractmethod
    def save(self, _dvc: dict) -> typing.Any:
        pass


class MockIndexdWriter(LoggingWriter):
    def save(self, dvc: DVC) -> str:
        self.logger.info(f'Saving {dvc}')
        return 'OK'


class IndexdWriter(LoggingWriter):
    """Submit a job to the indexd service, return response."""

    def __init__(self, log_file, auth: Gen3Auth, project_id: str, bucket_name: str, overwrite: bool, restricted_project_id: str, existing_ids: list[str], legacy_bucket: bool):
        super().__init__(log_file)
        self.auth = auth
        self.project_id = project_id
        self.bucket_name = bucket_name
        self.overwrite = overwrite
        self.restricted_project_id = restricted_project_id
        self.existing_ids = existing_ids
        self.legacy_bucket = legacy_bucket

    def save(self, dvc: DVC) -> str:
        from gen3_tracker.gen3.indexd import write_indexd
        self.logger.debug(f'Saving {dvc}')
        write_indexd(
            auth=self.auth,
            dvc=dvc,
            project_id=self.project_id,
            bucket_name=self.bucket_name,
            overwrite=self.overwrite,
            restricted_project_id=self.restricted_project_id,
            existing_records=self.existing_ids,
            legacy_bucket=self.legacy_bucket
        )
        return 'OK'


class MockJobWriter(LoggingWriter):
    """Submit a job to the submission service, return response."""

    def save(self, _dvc: dict) -> dict:
        self.logger.info(f'Saving {_dvc}')
        return {'job_info': ["TODO"]}  # {"job_info": ["TODO


class MockRemoteWriter(LoggingWriter):

    def __init__(self, log_file=None, remote=None):
        assert remote is not None, 'remote is required'
        super().__init__(log_file)
        self.remote = remote

    def save(self, _dvc: dict) -> str:
        self.logger.info(f'Saving to {self.remote} {_dvc}')
        return 'OK'


def to_manifest(dvc):
    return {
        'object_id': dvc.object_id,
        'md5': dvc.out.md5,
        'file_name': dvc.out.path,
        'size': dvc.out.size
    }


def to_s3(dvc):

    src = dvc.out.path
    if not pathlib.Path(src).exists():
        src = dvc.out.realpath
        if not src or not pathlib.Path(src).exists():
            src = dvc.out.source_url

    assert src, f"Could not determine source file. {dvc} "
    return f"aws s3 cp {src} s3://BUCKET-NAME/{dvc.object_id}/{dvc.out.path}"


class S3RemoteWriter(LoggingWriter):
    """Write files to S3."""

    def __init__(self, log_file=None, remote=None, work_dir=None):
        assert remote is not None, 'remote is required'
        super().__init__(log_file)
        self.remote = remote
        current_time = datetime.now().strftime("%Y%m%d%H%M%S")  # Format datetime as you need
        manifest_file = pathlib.Path(work_dir) / f'sh-manifest-{current_time}.sh'
        manifest_file.parent.mkdir(parents=True, exist_ok=True)
        self.manifest_file_path = manifest_file
        self.manifest = []

    def save(self, dvc: DVC) -> str:
        self.logger.info(f'Saving to S3 {dvc}')
        self.manifest.append(to_s3(dvc))
        return 'OK'

    def commit(self, dry_run=False, profile=None, upload_path=None, bucket_name=None, worker_count=(multiprocessing.cpu_count() - 1)):
        with open(self.manifest_file_path, 'w') as f:
            for _ in self.manifest:
                _ = _.replace('BUCKET-NAME', bucket_name)
                f.write(f"{_}\n")
        print(f'See manifest file: {self.manifest_file_path}')
        return 'OK'


class Gen3ClientRemoteWriter(LoggingWriter):
    """Write files to Gen3 using gen3-client."""

    def __init__(self, log_file=None, remote=None, work_dir=None):
        assert remote is not None, 'remote is required'
        super().__init__(log_file)
        self.remote = remote
        current_time = datetime.now().strftime("%Y%m%d%H%M%S")  # Format datetime as you need
        manifest_file = pathlib.Path(work_dir) / f'manifest-{current_time}.json'
        manifest_file.parent.mkdir(parents=True, exist_ok=True)
        self.manifest_file_path = manifest_file
        self.manifest = []

    def save(self, dvc: DVC) -> str:
        if dvc.out.realpath and not dvc.meta.no_bucket:
            self.logger.info(f'Saving to {self.remote} {dvc}')
            self.manifest.append(to_manifest(dvc))
        return 'OK'

    def commit(self, dry_run=False, profile=None, upload_path=None, bucket_name=None, worker_count=(multiprocessing.cpu_count() - 1)):
        with open(self.manifest_file_path, 'w') as f:
            json.dump(self.manifest, f)
        if len(self.manifest) > 0:
            cmd = f"gen3-client upload-multiple --manifest {self.manifest_file_path} --profile {profile} --upload-path {upload_path} --bucket {bucket_name} --numparallel {worker_count}"
            print(cmd)
            run_command(cmd, dry_run=dry_run, raise_on_err=True, no_capture=True)
        else:
            print(f'No files to upload to {self.remote} by gen3-client.')
        return 'OK'


def to_indexd(dvc_objects: list[DVC],
              auth: Gen3Auth,
              project_id: str,
              bucket_name: str,
              overwrite: bool,
              restricted_project_id: str,
              legacy_bucket: bool
              ) -> typing.Generator[typing.Any, None, None]:
    """Upload committed files to indexd."""
    # indexd_writer = MockIndexdWriter
    # log_file = "logs/mock-indexd.log"
    indexd_writer = IndexdWriter
    log_file = "logs/indexd.log"

    existing_ids = []
    for _ in dvc_objects:
        _.project_id = project_id
        existing_ids.append(_.object_id)

    with indexd_writer(auth=auth,
                       log_file=log_file,
                       project_id=project_id,
                       bucket_name=bucket_name,
                       overwrite=overwrite,
                       existing_ids=existing_ids,
                       restricted_project_id=restricted_project_id,
                       legacy_bucket=legacy_bucket) as indexd:
        for _ in dvc_objects:
            # add to indexd
            rc = indexd.save(_)
            yield rc


def to_remote(upload_method, dvc_objects, bucket_name, profile, dry_run, work_dir):
    """Upload committed files to remote."""
    # ['gen3', 's3', 's3-cp']
    if upload_method == 'gen3':
        writer = Gen3ClientRemoteWriter
    elif upload_method == 's3':
        writer = S3RemoteWriter
    else:
        logging.getLogger(__package__).info(f"No upload for {upload_method}")
        return

    with writer(work_dir=work_dir, log_file=f"logs/mock-remote-{upload_method}.log", remote=upload_method) as remote_writer:
        for _ in dvc_objects:
            remote_writer.save(_)
        remote_writer.commit(dry_run=dry_run, profile=profile, upload_path=pathlib.Path().cwd().resolve(), bucket_name=bucket_name)


def to_job(zip_file):
    """Upload zip files to submission"""
    with MockJobWriter(log_file="logs/mock-job.log") as job_writer:
        # TODO - add to remote
        job_writer.save(zip_file)
        time.sleep(3)




from gen3_tracker.collaborator.access.requestor import add_policies
from gen3_tracker.config import ensure_auth
from gen3_tracker.projects.lister import ls as project_ls
from typing import List


def initialize_project_server_side(config, project_id, auth=None) -> List[str] | bool:
    """Initialize a project in the current directory."""
    if auth is None:
        if not config.gen3.profile:
            return ["Disconnected mode, skipping server side initialization"]
        auth = ensure_auth(config=config)

    logs = []
    program, project = project_id.split('-')
    projects = project_ls(config, auth=auth)
    existing_project = [_ for _ in projects.complete if _.endswith(f"{program}-{project}")]
    policy_msgs = []
    # looking for prject to exist, but not reader permissions on project, then sign project.
    # If at least reader permissions then carry on.
    if len(existing_project) > 0:
        logs.append(f"Pending request for project creation. Admin must sign access request for: /{program}/projects/{project}")
        return logs, False
    else:
        _ = add_policies(config, project_id, auth=auth)
        policy_msgs.extend([_.msg, f"See {_.commands}"])
        logs.extend(policy_msgs)
    return logs, True



import asyncio

from gen3.jobs import Gen3Jobs

from gen3_tracker.config import Config, ensure_auth


def empty(config: Config, project_id: str, wait: bool = True) -> dict:
    """Empty all meta data (graph, flat) for a project."""

    assert '-' in project_id, f'Invalid project_id: {project_id}'
    program, project = project_id.split('-')
    assert program and project, f'Invalid project_id: {project_id}'

    auth = ensure_auth(config=config)
    jobs_client = Gen3Jobs(auth_provider=auth)
    args = {'object_id': None, 'project_id': project_id, 'method': 'delete'}

    if wait:
        _ = asyncio.run(jobs_client.async_run_job_and_wait('fhir_import_export', args))
    else:
        _ = jobs_client.create_job('fhir_import_export', args)
        _ = {'output': _}
    return _



import logging
import sys

import click

import gen3_tracker
from gen3_tracker import NaturalOrderGroup, ENV_VARIABLE_PREFIX
from gen3_tracker.collaborator.cli import collaborator
from gen3_tracker.config import gen3_client_profiles
from gen3_tracker.git.cli import cli as git
from gen3_tracker.meta.cli import meta
from gen3_tracker.projects.cli import project_group as project

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__package__)


@click.group(cls=NaturalOrderGroup)
@click.option('--format', 'output_format',
              envvar=f"{ENV_VARIABLE_PREFIX}FORMAT",
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json', 'text'], case_sensitive=False),
              help=f'Result format. {ENV_VARIABLE_PREFIX}FORMAT'
              )
@click.option('--profile', 'profile',
              envvar=f"{ENV_VARIABLE_PREFIX}PROFILE",
              default=None,
              show_default=True,
              help=f'Connection name. {ENV_VARIABLE_PREFIX}PROFILE See https://bit.ly/3NbKGi4'
              )
@click.option('--debug', is_flag=True, envvar='G3T_DEBUG', help='Enable debug mode. G3T_DEBUG environment variable can also be used.')
@click.option('--dry-run', is_flag=True, envvar='G3T_DRYRUN', help='Print the commands that would be executed, but do not execute them. G3T_DRYRUN environment variable can also be used.')
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, output_format: str, profile: str, debug: bool, dry_run: bool):
    """A CLI for adding version control to Gen3 projects."""
    config__ = gen3_tracker.config.default()
    logging.basicConfig(format=config__.log.format, level=config__.log.level, stream=sys.stderr)

    if output_format:
        config__.output.format = output_format

    _profiles = gen3_client_profiles()
    is_help = '--help' in sys.argv[1:]

    if profile:
        if profile not in _profiles:
            click.secho(f"Profile {profile} not found.", fg='red')
            exit(1)
        config__.gen3.profile = profile
    elif not config__.gen3.profile and not is_help:
        if not _profiles:
            click.secho("No gen3_client profile found.", fg='red')
            exit(1)
        else:
            if len(_profiles) > 1:
                click.secho(f"WARNING: No --profile specified, found multiple gen3_client profiles: {_profiles}",
                            fg='red')
            else:
                click.secho(f"Using default gen3_client profile {_profiles[0]}", fg='yellow')
                config__.gen3.profile = _profiles[0]

    # ensure that ctx.obj exists
    config__.debug = debug
    config__.dry_run = dry_run

    ctx.obj = config__

    if debug:
        _logger.setLevel(logging.DEBUG)


git.add_command(meta)
git.add_command(collaborator)
git.add_command(project)

for command in git.commands.values():
    cli.add_command(command)


if __name__ == "__main__":
    cli()
